from typing import Optional, Mapping, overload

import pulumi
from pulumi_aws import backup, get_caller_identity
from pulumi_aws.backup import VaultArgs
from cloudlib.modules.resource import BaseFactory
from cloudlib.modules.kms import KeyFactory


class VaultFactory(BaseFactory):
    def __init__(
        self,
        name: Optional[pulumi.Input[str]],
        opts: Optional[pulumi.ResourceOptions] = None,
        force_destroy: Optional[pulumi.Input[bool]] = None,
        kms_key_arn: Optional[pulumi.Input[str]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        backup_stack=None,
    ):
        name = "Staging" if name == "staging" else name
        super().__init__(
            resource_name=name, opts=opts, force_destroy=force_destroy, kms_key_arn=kms_key_arn, name=name, tags=tags
        )
        self.create_default_plan = False
        self.backup_stack = backup_stack

    def with_default_plan(self):
        self.create_default_plan = True

    def _create_default_plan(self):
        vault = "Staging" if self.name == "staging" else self.name

        identity = get_caller_identity()
        plan = backup.Plan(
            self.resource_name,
            name=vault,
            rules=[
                backup.PlanRuleArgs(
                    completion_window=10080,
                    enable_continuous_backup=True,
                    lifecycle=backup.PlanRuleLifecycleArgs(
                        delete_after=35,
                    ),
                    copy_actions=[
                        backup.PlanRuleCopyActionArgs(
                            destination_vault_arn=self.backup_stack.get_output("copy_vault"),
                            lifecycle=backup.PlanRuleCopyActionLifecycleArgs(
                                delete_after=21,
                            ),
                        )
                    ],
                    rule_name="DailyBackups",
                    schedule="cron(0 5 ? * * *)",
                    start_window=480,
                    target_vault_name=vault,
                ),
                backup.PlanRuleArgs(
                    completion_window=10080,
                    lifecycle=backup.PlanRuleLifecycleArgs(
                        delete_after=90,
                    ),
                    copy_actions=[
                        backup.PlanRuleCopyActionArgs(
                            destination_vault_arn=self.backup_stack.get_output("copy_vault"),
                            lifecycle=backup.PlanRuleCopyActionLifecycleArgs(
                                delete_after=180,
                            ),
                        )
                    ],
                    rule_name="WeeklyBackups",
                    schedule="cron(0 5 ? * 7 *)",
                    start_window=480,
                    target_vault_name=vault,
                ),
                backup.PlanRuleArgs(
                    completion_window=10080,
                    lifecycle=backup.PlanRuleLifecycleArgs(
                        cold_storage_after=90,
                        delete_after=1825,
                    ),
                    rule_name="MonthlyBackups",
                    schedule="cron(0 5 1 * ? *)",
                    start_window=480,
                    target_vault_name=vault,
                ),
            ],
        )

        backup.Selection(
            vault,
            conditions=[backup.SelectionConditionArgs()],
            iam_role_arn=f"arn:aws:iam::{identity.account_id}:role/service-role/AWSBackupDefaultServiceRole",
            name="All",
            plan_id=plan.id,
            resources=[
                "arn:aws:rds:*:*:cluster:*",
                "arn:aws:timestream:*:*:database/*",
                "arn:aws:s3:::*",
                "arn:aws:dynamodb:*:*:table/*",
                "arn:aws:rds:*:*:db:*",
                "arn:aws:cloudformation:*:*:stack/*",
            ],
        )

    def create(self):
        if not self.kwargs.get("kms_key_arn"):
            key_factory = KeyFactory("backup")
            key_factory.with_alias("backup")
            key = key_factory.create()

            self.setdefault("kms_key_arn", key.arn)

        vault = self._create(backup.Vault)

        if self.create_default_plan:
            self._create_default_plan()

        return vault
