from typing import Mapping, Optional, Sequence

import pulumi
from pulumi_aws import ec2

from cloudlib.modules.resource import BaseFactory


class SecurityGroupFactory(BaseFactory):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        description: Optional[pulumi.Input[str]] = None,
        egress: Optional[pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["SecurityGroupEgressArgs"]]]]] = None,
        ingress: Optional[pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["SecurityGroupIngressArgs"]]]]] = None,
        name_prefix: Optional[pulumi.Input[str]] = None,
        revoke_rules_on_delete: Optional[pulumi.Input[bool]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        vpc_id: Optional[pulumi.Input[str]] = None,
    ):
        super().__init__(
            resource_name=name,
            name=name,
            opts=opts,
            description=description,
            egress=egress,
            ingress=ingress,
            name_prefix=name_prefix,
            revoke_rules_on_delete=revoke_rules_on_delete,
            tags=tags,
            vpc_id=vpc_id,
        )

        self.ingress_rules = []
        self.egress_rules = []

    def with_ingress(
        self,
        from_port: int,
        protocol: str,
        to_port: int,
        cidr_blocks: Sequence[str] = None,
        description: str = None,
        ipv6_cidr_blocks: Sequence[str] = None,
        prefix_list_ids: Sequence[str] = None,
        security_groups: Sequence[str] = None,
    ):
        self.ingress_rules.append(
            ec2.SecurityGroupIngressArgs(
                from_port=from_port,
                protocol=protocol,
                to_port=to_port,
                cidr_blocks=cidr_blocks,
                description=description,
                ipv6_cidr_blocks=ipv6_cidr_blocks,
                prefix_list_ids=prefix_list_ids,
                security_groups=security_groups,
            )
        )

    def with_egress(
        self,
        from_port: int,
        protocol: str,
        to_port: int,
        cidr_blocks: Sequence[str] = None,
        description: str = None,
        ipv6_cidr_blocks: Sequence[str] = None,
        prefix_list_ids: Sequence[str] = None,
        security_groups: Sequence[str] = None,
    ):
        self.egress_rules.append(
            ec2.SecurityGroupIngressArgs(
                from_port=from_port,
                protocol=protocol,
                to_port=to_port,
                cidr_blocks=cidr_blocks,
                description=description,
                ipv6_cidr_blocks=ipv6_cidr_blocks,
                prefix_list_ids=prefix_list_ids,
                security_groups=security_groups,
            )
        )

    def with_https_egress(self, cidr_blocks=None, security_groups=None, **kwargs):
        self.with_egress(
            from_port=443,
            to_port=443,
            protocol="tcp",
            cidr_blocks=cidr_blocks,
            security_groups=security_groups,
            **kwargs
        )

    def create(self) -> ec2.SecurityGroup:
        self.kwargs["egress"] = (self.kwargs.get("egress") or []) + self.egress_rules
        self.kwargs["ingress"] = (self.kwargs.get("ingress") or []) + self.ingress_rules

        return self._create(ec2.SecurityGroup)
