import pulumi
from typing import Optional, Mapping
from pulumi_aws import sns, get_caller_identity

from cloudlib.modules.cloudwatch import LogGroupFactory
from cloudlib.modules.kms import KeyFactory
from cloudlib.modules.resource import BaseFactory


class TopicSubscriptionFactory(BaseFactory):
    def __init__(
        self,
        opts: Optional[pulumi.ResourceOptions] = None,
        confirmation_timeout_in_minutes: Optional[pulumi.Input[int]] = None,
        delivery_policy: Optional[pulumi.Input[str]] = None,
        endpoint: Optional[pulumi.Input[str]] = None,
        endpoint_auto_confirms: Optional[pulumi.Input[bool]] = None,
        filter_policy: Optional[pulumi.Input[str]] = None,
        filter_policy_scope: Optional[pulumi.Input[str]] = None,
        protocol: Optional[pulumi.Input[str]] = None,
        raw_message_delivery: Optional[pulumi.Input[bool]] = None,
        redrive_policy: Optional[pulumi.Input[str]] = None,
        subscription_role_arn: Optional[pulumi.Input[str]] = None,
        topic: Optional[pulumi.Input[str]] = None,
        resource_name=None,
    ):
        super().__init__(
            resource_name=(resource_name or endpoint),
            opts=opts,
            confirmation_timeout_in_minutes=confirmation_timeout_in_minutes,
            delivery_policy=delivery_policy,
            endpoint=endpoint,
            endpoint_auto_confirms=endpoint_auto_confirms,
            filter_policy=filter_policy,
            filter_policy_scope=filter_policy_scope,
            protocol=protocol,
            raw_message_delivery=raw_message_delivery,
            redrive_policy=redrive_policy,
            subscription_role_arn=subscription_role_arn,
            topic=topic,
        )

    def create(self):
        return self._create(sns.TopicSubscription)


class TopicFactory(BaseFactory):
    def __init__(
        self,
        name: Optional[pulumi.Input[str]],
        opts: Optional[pulumi.ResourceOptions] = None,
        application_failure_feedback_role_arn: Optional[pulumi.Input[str]] = None,
        application_success_feedback_role_arn: Optional[pulumi.Input[str]] = None,
        application_success_feedback_sample_rate: Optional[pulumi.Input[int]] = None,
        content_based_deduplication: Optional[pulumi.Input[bool]] = None,
        delivery_policy: Optional[pulumi.Input[str]] = None,
        display_name: Optional[pulumi.Input[str]] = None,
        fifo_topic: Optional[pulumi.Input[bool]] = None,
        firehose_failure_feedback_role_arn: Optional[pulumi.Input[str]] = None,
        firehose_success_feedback_role_arn: Optional[pulumi.Input[str]] = None,
        firehose_success_feedback_sample_rate: Optional[pulumi.Input[int]] = None,
        http_failure_feedback_role_arn: Optional[pulumi.Input[str]] = None,
        http_success_feedback_role_arn: Optional[pulumi.Input[str]] = None,
        http_success_feedback_sample_rate: Optional[pulumi.Input[int]] = None,
        kms_master_key_id: Optional[pulumi.Input[str]] = None,
        lambda_failure_feedback_role_arn: Optional[pulumi.Input[str]] = None,
        lambda_success_feedback_role_arn: Optional[pulumi.Input[str]] = None,
        lambda_success_feedback_sample_rate: Optional[pulumi.Input[int]] = None,
        name_prefix: Optional[pulumi.Input[str]] = None,
        policy: Optional[pulumi.Input[str]] = None,
        signature_version: Optional[pulumi.Input[int]] = None,
        sqs_failure_feedback_role_arn: Optional[pulumi.Input[str]] = None,
        sqs_success_feedback_role_arn: Optional[pulumi.Input[str]] = None,
        sqs_success_feedback_sample_rate: Optional[pulumi.Input[int]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        tracing_config: Optional[pulumi.Input[str]] = None,
    ):
        super().__init__(
            resource_name=name,
            opts=opts,
            application_failure_feedback_role_arn=application_failure_feedback_role_arn,
            application_success_feedback_role_arn=application_success_feedback_role_arn,
            application_success_feedback_sample_rate=application_success_feedback_sample_rate,
            content_based_deduplication=content_based_deduplication,
            delivery_policy=delivery_policy,
            display_name=display_name,
            fifo_topic=fifo_topic,
            firehose_failure_feedback_role_arn=firehose_failure_feedback_role_arn,
            firehose_success_feedback_role_arn=firehose_success_feedback_role_arn,
            firehose_success_feedback_sample_rate=firehose_success_feedback_sample_rate,
            http_failure_feedback_role_arn=http_failure_feedback_role_arn,
            http_success_feedback_role_arn=http_success_feedback_role_arn,
            http_success_feedback_sample_rate=http_success_feedback_sample_rate,
            kms_master_key_id=kms_master_key_id,
            lambda_failure_feedback_role_arn=lambda_failure_feedback_role_arn,
            lambda_success_feedback_role_arn=lambda_success_feedback_role_arn,
            lambda_success_feedback_sample_rate=lambda_success_feedback_sample_rate,
            name=name,
            name_prefix=name_prefix,
            policy=policy,
            signature_version=signature_version,
            sqs_failure_feedback_role_arn=sqs_failure_feedback_role_arn,
            sqs_success_feedback_role_arn=sqs_success_feedback_role_arn,
            sqs_success_feedback_sample_rate=sqs_success_feedback_sample_rate,
            tags=tags,
            tracing_config=tracing_config,
        )
        self.subscriptions = []
        self.logging = None

    def with_https_subscription(self, target, **kwargs):
        kwargs.setdefault("protocol", "https")
        identity = get_caller_identity()
        self.setdefault("http_failure_feedback_role_arn", f"arn:aws:iam::{identity.account_id}:role/SnsFailureFeedback")
        self.setdefault("http_success_feedback_role_arn", f"arn:aws:iam::{identity.account_id}:role/SnsSuccessFeedback")
        self.setdefault("http_success_feedback_sample_rate", "10")
        self.with_subscription(endpoint=target, **kwargs)

    def with_subscription(self, **kwargs):
        self.subscriptions.append(TopicSubscriptionFactory(**kwargs))

    def create(self):
        if not self.kwargs.get("kms_master_key_id"):
            key_factory = KeyFactory(self.name)
            key_factory.with_alias(self.name)
            key = key_factory.create()

            self.setdefault("kms_master_key_id", key.arn)

        topic = self._create(sns.Topic)

        for sub in self.subscriptions:
            sub.setdefault("topic", topic.arn)
            sub.create()

        return topic
