from typing import Optional, Sequence, Mapping

from pulumi_aws import wafv2, get_caller_identity, iam, get_region

from cloudlib.modules.resource import BaseFactory
from cloudlib.modules.s3 import BucketPolicyFactory
from cloudlib.utils.names import IdentifierV2


class WafS3PolicyFactory(BucketPolicyFactory):
    def __init__(self, accounts=None):
        self.accounts = accounts
        super().__init__()

    def create(self, bucket):
        arn = f"arn:aws:s3:::{bucket}"

        current_caller_identity = get_caller_identity()
        region = get_region()

        if not self.accounts:
            self.accounts = [current_caller_identity.account_id]

        statements = []
        for account_id in self.accounts:
            statements.append(
                iam.GetPolicyDocumentStatementArgs(
                    sid=f"AWSLogDeliveryAclCheck_{account_id}",
                    effect="Allow",
                    principals=[
                        iam.GetPolicyDocumentStatementPrincipalArgs(
                            type="Service",
                            identifiers=["delivery.logs.amazonaws.com"],
                        )
                    ],
                    actions=["s3:GetBucketAcl"],
                    resources=[arn],
                    conditions=[
                        iam.GetPolicyDocumentStatementConditionArgs(
                            test="StringEquals",
                            variable="aws:SourceAccount",
                            values=[account_id],
                        ),
                        iam.GetPolicyDocumentStatementConditionArgs(
                            test="ArnLike",
                            variable="AWS:SourceArn",
                            values=[f"arn:aws:logs:*:{account_id}:*"],
                        ),
                    ],
                )
            )

            statements.append(
                iam.GetPolicyDocumentStatementArgs(
                    sid=f"AWSLogDeliveryWrite_{account_id}",
                    effect="Allow",
                    principals=[
                        iam.GetPolicyDocumentStatementPrincipalArgs(
                            type="Service",
                            identifiers=["delivery.logs.amazonaws.com"],
                        )
                    ],
                    actions=["s3:PutObject"],
                    resources=[f"{arn}/AWSLogs/{account_id}/*"],
                    conditions=[
                        iam.GetPolicyDocumentStatementConditionArgs(
                            test="StringEquals",
                            variable="s3:x-amz-acl",
                            values=["bucket-owner-full-control"],
                        ),
                        iam.GetPolicyDocumentStatementConditionArgs(
                            test="StringEquals",
                            variable="aws:SourceAccount",
                            values=[account_id],
                        ),
                        iam.GetPolicyDocumentStatementConditionArgs(
                            test="ArnLike",
                            variable="AWS:SourceArn",
                            values=[f"arn:aws:logs:us-east-1:{account_id}:*"],
                        ),
                    ],
                )
            )

        return super().create(bucket, extra_statements=statements)


class WebAclFactory(BaseFactory):
    def __init__(
        self,
        name: str,
        opts: Optional["ResourceOptions"] = None,
        association_config: Optional["WebAclAssociationConfigArgs"] = None,
        captcha_config: Optional["WebAclCaptchaConfigArgs"] = None,
        custom_response_bodies: Optional[Sequence["WebAclCustomResponseBodyArgs"]] = None,
        default_action: Optional["WebAclDefaultActionArgs"] = None,
        description: Optional[str] = None,
        rules: Optional[Sequence["WebAclRuleArgs"]] = None,
        scope: Optional[str] = None,
        tags: Optional[Mapping[str, str]] = None,
        token_domains: Optional[Sequence[str]] = None,
        visibility_config: Optional["WebAclVisibilityConfigArgs"] = None,
    ):
        super().__init__(
            resource_name=name,
            name=name,
            opts=opts,
            association_config=association_config,
            captcha_config=captcha_config,
            custom_response_bodies=custom_response_bodies,
            default_action=default_action,
            description=description,
            rules=rules or [],
            scope=scope,
            tags=tags,
            token_domains=token_domains,
            visibility_config=visibility_config,
        )

    def with_rule(
        self,
        name,
        statement,
        visibility_config,
        priority: int = None,
        action=None,
        captcha_config=None,
        override_action=None,
        rule_labels=None,
    ):
        self.kwargs["rules"].append(
            wafv2.WebAclRuleArgs(
                name=name,
                priority=priority or (len(self.kwargs["rules"]) + 1) * 100,
                statement=statement,
                visibility_config=visibility_config,
                action=action,
                captcha_config=captcha_config,
                override_action=override_action,
                rule_labels=rule_labels,
            )
        )

    def with_ipset(self, name, ip_addresses):
        return wafv2.IpSet(
            IdentifierV2(name).snake,
            addresses=ip_addresses,
            description=IdentifierV2(f"{name} IP set").spinal,
            ip_address_version="IPV4",
            scope=self.kwargs.get("scope"),
        )

        # def with_rule_group(self, name, *rules, **kwargs):
        #     kwargs.setdefault("scope", self.kwargs.get("scope"))
        #     kwargs["name"] = name
        #     kwargs["rules"] = rules

        # kwargs.setdefault("visibility_config", {
        #     "SampledRequestsEnabled": True,
        #     "CloudWatchMetricsEnabled": True,
        #     "MetricName": name
        # })
        # opts: Optional[ResourceOptions] = None,
        # capacity: Optional[int] = None,
        # custom_response_bodies: Optional[Sequence[RuleGroupCustomResponseBodyArgs]] = None,
        # description: Optional[str] = None,
        # name_prefix: Optional[str] = None,
        # rules: Optional[Sequence[RuleGroupRuleArgs]] = None,

        wafv2.RuleGroup(resource_name=IdentifierV2(name).snake, **kwargs)

    def with_managed_rule(
        self,
        name,
        vendor,
        rule_name,
        visibility_config=None,
        action=None,
        priority: int = None,
        managed_rule_group_configs=None,
    ):
        self.with_rule(
            name=name,
            priority=priority,
            override_action=wafv2.WebAclRuleOverrideActionArgs(
                none=wafv2.WebAclRuleOverrideActionNoneArgs(),
            ),
            statement=wafv2.WebAclRuleStatementArgs(
                managed_rule_group_statement=wafv2.WebAclRuleStatementManagedRuleGroupStatementArgs(
                    managed_rule_group_configs=managed_rule_group_configs or [],
                    name=rule_name,
                    # rule_action_overrides=[wafv2.WebAclRuleStatementManagedRuleGroupStatementRuleActionOverrideArgs(
                    #     action_to_use=wafv2.WebAclRuleStatementManagedRuleGroupStatementRuleActionOverrideActionToUseArgs(
                    #         count=wafv2.WebAclRuleStatementManagedRuleGroupStatementRuleActionOverrideActionToUseCountArgs()
                    #     ),
                    #     name="SizeRestrictions_BODY",
                    # )],
                    vendor_name=vendor,
                ),
            ),
            visibility_config=visibility_config
            or wafv2.WebAclVisibilityConfigArgs(
                cloudwatch_metrics_enabled=True, metric_name=f"WebACL-{self.name}-{name}", sampled_requests_enabled=True
            ),
            action=action,
        )
        self.logging_target = None

    def create(self):
        self.setdefault("default_action", wafv2.WebAclDefaultActionArgs(allow=wafv2.WebAclDefaultActionAllowArgs()))
        self.setdefault("scope", "REGIONAL")
        self.setdefault(
            "visibility_config",
            wafv2.WebAclVisibilityConfigArgs(
                cloudwatch_metrics_enabled=True, metric_name=f"WebACL-{self.name}", sampled_requests_enabled=True
            ),
        )

        web_acl = self._create(wafv2.WebAcl)

        if self.logging_target:
            wafv2.WebAclLoggingConfiguration(
                IdentifierV2(self.resource_name),
                log_destination_configs=[self.logging_target],
                resource_arn=web_acl.arn,
                redacted_fields=[
                    wafv2.WebAclLoggingConfigurationRedactedFieldArgs(
                        method=wafv2.WebAclLoggingConfigurationRedactedFieldMethodArgs(),
                    ),
                    wafv2.WebAclLoggingConfigurationRedactedFieldArgs(
                        query_string=wafv2.WebAclLoggingConfigurationRedactedFieldQueryStringArgs(),
                    ),
                    wafv2.WebAclLoggingConfigurationRedactedFieldArgs(
                        uri_path=wafv2.WebAclLoggingConfigurationRedactedFieldUriPathArgs(),
                    ),
                    wafv2.WebAclLoggingConfigurationRedactedFieldArgs(
                        single_header=wafv2.WebAclLoggingConfigurationRedactedFieldSingleHeaderArgs(
                            name="user-agent",
                        ),
                    ),
                ],
            )

        return web_acl

    def with_logging(self, logging_target):
        self.logging_target = logging_target
