# Networking model for AWS Cloud Rules
# 1. each stage should have it's own /10 IPv4 block assigned to it.
# 2. /10 is big enough to handle whole workload we need, if in future we will need to accomodate more services etc,
# we should consider another instance of the cloud to mitigate risk
# 3. each region is getting it's own /14 block adress - that will handle up to 16 regions in one stage, again if we need
# more, we should create second instance of the cloud
# 4. each region should define list of apps available in given region and assign dedicated /19 IP block, that will give
# us possibility to register up to 30 isolated apps in one region (32 ip blocks - 2 reserved for networking)
# 5. each region will have it's own network setup in which first 2 /19 IP blocks are reserved and used as 4 /20 networks
# where 1st /20 block is used for core-ingress network, 2nd /20 is for core-egress network, 3rd /20 block is for inspection network
# and 4th /20 is used by transit gateway as a interconnection network (each vpc will get /28 from that block)
from collections import OrderedDict
from typing import List
from pulumi_aws import get_availability_zones

from netaddr import IPNetwork

apps = ["epsy-app", "portal-app", "admin-app", "vta-app", "clinician-app"]

network = dict(
    prod=dict(
        cidr_block="10.64.0.0/10",
        regions=OrderedDict(
            {
                "us-east-1": dict(apps=apps, az=2),
                "eu-west-2": dict(apps=apps, az=2),
            }
        ),
    ),
    staging=dict(
        cidr_block="10.128.0.0/10",
        regions=OrderedDict(
            {
                "us-east-1": dict(apps=apps, az=2),
                "eu-west-2": dict(apps=apps, az=2),
            }
        ),
    ),
    dev=dict(
        cidr_block="10.192.0.0/10",
        regions=OrderedDict(
            {
                "us-east-1": dict(apps=apps, az=2),
                "eu-west-2": dict(apps=apps, az=2),
            }
        ),
    ),
    play=dict(
        cidr_block="10.0.0.0/10",
        regions=OrderedDict(
            {
                "us-east-1": dict(apps=["play"], az=1),
            }
        ),
    ),
)


class AppNetwork:
    def __init__(self, app_cidr: IPNetwork, shared_cidr: IPNetwork, az: List, az_ids: List):
        self.app_cidr = app_cidr
        self.shared_cidr = shared_cidr
        self.az = az
        self.az_ids = az_ids
        self.firewall_network, self.transit_network = list(self.shared_cidr.subnet(27))

    def get_firewall_network(self):
        return self.firewall_network

    def get_transit_network(self):
        return self.transit_network


class RegionNetwork:
    def __init__(self, network_cidr: IPNetwork, az, shared_block: IPNetwork, apps: List, name=None):
        self.network_cidr = network_cidr
        self.name = name
        _shared = list(shared_block.subnet(26))
        self.app_shared_blocks = dict(zip(apps, _shared[4 : len(apps) + 4]))

        _subnets = list(self.network_cidr.subnet(19))
        _reserved = list(_subnets[0].subnet(20)) + list(_subnets[1].subnet(20))

        self.az = az
        availability_zones = get_availability_zones().zone_ids
        availability_zones.sort()
        self.az_ids = availability_zones[0 : self.az]

        self.ingress_network = AppNetwork(app_cidr=_reserved[0], shared_cidr=_shared[0], az=az, az_ids=self.az_ids)
        self.egress_network = AppNetwork(app_cidr=_reserved[1], shared_cidr=_shared[1], az=az, az_ids=self.az_ids)
        self.inspect_network = AppNetwork(app_cidr=_reserved[2], shared_cidr=_shared[2], az=az, az_ids=self.az_ids)
        self.shared_network = AppNetwork(app_cidr=_reserved[3], shared_cidr=_shared[3], az=az, az_ids=self.az_ids)

        self.app_networks = dict(zip(apps, _subnets[2 : len(apps) + 2]))

        self.shared_block = shared_block
        self.apps = {
            name: AppNetwork(
                app_cidr=self.app_networks[name],
                shared_cidr=self.app_shared_blocks[name],
                az=self.az,
                az_ids=self.az_ids,
            )
            for name in apps
        }

    def get_app_network(self, name) -> AppNetwork:
        return self.apps[name]


class Region:
    def __init__(self, stage, name, network_cidr, apps, az, shared_block):
        self.stage = stage
        self.name = name
        self.apps = apps
        self.network = RegionNetwork(
            network_cidr=network_cidr, az=az, shared_block=shared_block, apps=apps, name=self.name
        )

    @staticmethod
    def get(stage, name) -> "Region":
        stage_network = IPNetwork(network[stage]["cidr_block"])
        regions = network[stage]["regions"]
        region_networks = dict(zip(regions, list(stage_network.subnet(14))[0 : len(regions)]))

        shared = list(IPNetwork("100.64.0.0/10").subnet(15))
        shared_block = shared[list(regions.keys()).index(name)]

        return Region(
            stage=stage,
            name=name,
            network_cidr=region_networks.get(name),
            **network[stage]["regions"][name],
            shared_block=shared_block
        )
