from typing import Optional, Sequence, Mapping, overload

import pulumi
from pulumi_aws import acm, route53
from pulumi_aws.acm import CertificateArgs

from cloudlib.modules.resource import BaseFactory
from cloudlib.utils.names import Identifier


def certificate(fqdn, alternative_names=None, zone=None, zone_id=None, prefix="global") -> acm.Certificate:
    namespace = Identifier(prefix).add(fqdn)

    cert = acm.Certificate(
        resource_name=namespace,
        domain_name=fqdn,
        validation_method="DNS",
        subject_alternative_names=alternative_names or [],
    )

    if isinstance(zone, str):
        validation_zone = route53.get_zone(name=zone, private_zone=False)
    elif zone_id:
        validation_zone = route53.get_zone(zone_id=zone_id, private_zone=False)
    else:
        validation_zone = zone

    def build_validation_record(validation_option):
        records = []

        for idx, option in enumerate(validation_option):
            route53.Record(
                namespace.add(f"{option.resource_record_name}_{idx}_validation"),
                ttl=60,
                zone_id=validation_zone.zone_id,
                allow_overwrite=True,
                name=option.resource_record_name,
                records=[option.resource_record_value],
                type=route53.RecordType(option.resource_record_type),
            )

        return records

    validation = cert.domain_validation_options.apply(build_validation_record)

    acm.CertificateValidation(
        resource_name=namespace.add("validation"), certificate_arn=cert.arn, validation_record_fqdns=validation
    )

    return cert


class CertificateFactory(BaseFactory):
    def __init__(
        self,
        resource_name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        certificate_authority_arn: Optional[pulumi.Input[str]] = None,
        certificate_body: Optional[pulumi.Input[str]] = None,
        certificate_chain: Optional[pulumi.Input[str]] = None,
        domain_name: Optional[pulumi.Input[str]] = None,
        early_renewal_duration: Optional[pulumi.Input[str]] = None,
        key_algorithm: Optional[pulumi.Input[str]] = None,
        options: Optional[pulumi.Input[pulumi.InputType["CertificateOptionsArgs"]]] = None,
        private_key: Optional[pulumi.Input[str]] = None,
        subject_alternative_names: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        validation_method: Optional[pulumi.Input[str]] = None,
        validation_options: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["CertificateValidationOptionArgs"]]]]
        ] = None,
    ):
        super().__init__(
            resource_name=resource_name,
            opts=opts,
            certificate_authority_arn=certificate_authority_arn,
            certificate_body=certificate_body,
            certificate_chain=certificate_chain,
            domain_name=domain_name,
            early_renewal_duration=early_renewal_duration,
            key_algorithm=key_algorithm,
            options=options,
            private_key=private_key,
            subject_alternative_names=subject_alternative_names,
            tags=tags,
            validation_method=validation_method,
            validation_options=validation_options,
        )

    def create(self):
        self.set_default("validation_method", "DNS")
        return self._create(acm.Certificate)
