from pulumi_aws import acmpca
from pulumi_aws_native import acmpca as acmpca_native

from cloudlib.utils.names import Identifier
from cloudlib.utils.pulumi import depends_on


class RootCertificateAuthority(acmpca_native.CertificateAuthority):
    def __init__(__self__, resource_name: str, *args, **kwargs):
        kwargs.setdefault("key_algorithm", "RSA_4096")
        kwargs.setdefault("signing_algorithm", "SHA512WITHRSA")
        kwargs.setdefault(
            "revocation_configuration",
            acmpca.CertificateAuthorityRevocationConfigurationArgs(
                ocsp_configuration=acmpca.CertificateAuthorityRevocationConfigurationOcspConfigurationArgs(enabled=True)
            ),
        )
        kwargs["type"] = "ROOT"
        super().__init__(resource_name, *args, **kwargs)


class SubordinateAuthority(acmpca_native.CertificateAuthority):
    def __init__(__self__, resource_name: str, *args, **kwargs):
        kwargs.setdefault("key_algorithm", "RSA_2048")
        kwargs.setdefault("signing_algorithm", "SHA256WITHRSA")
        kwargs["type"] = "SUBORDINATE"
        super().__init__(resource_name, *args, **kwargs)


class Certificate(acmpca_native.Certificate):
    def __init__(__self__, resource_name: str, ca, certificate, *args, **kwargs):
        kwargs.setdefault("certificate_authority_arn", ca.arn)
        kwargs.setdefault("certificate_signing_request", certificate.certificate_signing_request)
        kwargs.setdefault("signing_algorithm", "SHA256WITHRSA")
        super().__init__(resource_name, *args, **kwargs)


class CertificateL1(Certificate):
    def __init__(__self__, resource_name: str, ca, *args, **kwargs):
        kwargs.setdefault("template_arn", "arn:aws:acm-pca:::template/RootCACertificate/V1")
        kwargs.setdefault("validity", acmpca.CertificateValidityArgs(type="DAYS", value=3650))
        super().__init__(resource_name, ca=ca, certificate=ca, *args, **kwargs)


class CertificateL2(Certificate):
    def __init__(__self__, resource_name: str, ca, certificate, *args, **kwargs):
        kwargs.setdefault("template_arn", "arn:aws:acm-pca:::template/SubordinateCACertificate_PathLen3/V1")
        kwargs.setdefault("validity", acmpca.CertificateValidityArgs(type="DAYS", value=1095))
        super().__init__(resource_name, ca=ca, certificate=certificate, *args, **kwargs)


class CertificateL3(Certificate):
    def __init__(__self__, resource_name: str, ca, certificate, *args, **kwargs):
        kwargs.setdefault(
            "template_arn",
            "arn:aws:acm-pca:::template/SubordinateCACertificate_PathLen2/V1",
        )
        kwargs.setdefault("validity", acmpca.CertificateValidityArgs(type="DAYS", value=730))
        super().__init__(resource_name, ca=ca, certificate=certificate, *args, **kwargs)


class Activation(acmpca_native.CertificateAuthorityActivation):
    def __init__(__self__, resource_name: str, cert, ca, *args, **kwargs):
        kwargs.setdefault("status", "ACTIVE")
        kwargs.setdefault("certificate", cert.certificate)
        kwargs.setdefault("certificate_authority_arn", ca.arn)

        super().__init__(resource_name, *args, **kwargs)


class DefaultPermissions(acmpca_native.Permission):
    def __init__(__self__, cert, resource_name: str, *args, **kwargs):
        kwargs.setdefault("actions", ["IssueCertificate", "GetCertificate", "ListPermissions"])
        kwargs.setdefault("principal", "acm.amazonaws.com")
        kwargs.setdefault("certificate_authority_arn", cert.arn)
        super().__init__(resource_name, *args, **kwargs)


def certificate_authority_tree(country, organization, common_name, products):
    identifier = Identifier("certificate_authority_" + common_name)

    certs = {}

    base_subject = dict(
        country=country,
        organization=organization,
    )

    ca_id = identifier.add("ca")
    ca = RootCertificateAuthority(
        resource_name=ca_id, subject=acmpca_native.CertificateAuthoritySubjectArgs(**base_subject)
    )

    l1 = ca_id.add("cert")
    certificate_l1 = CertificateL1(resource_name=l1, ca=ca, opts=depends_on(ca))

    root_activation = Activation(
        resource_name=l1.add("validation"), cert=certificate_l1, ca=ca, opts=depends_on(ca, certificate_l1)
    )
    DefaultPermissions(
        resource_name=l1.add("permission"), cert=ca, opts=depends_on(ca, certificate_l1, root_activation)
    )
    certs["root"] = root_activation.complete_certificate_chain

    for group, apps in products.items():
        sub_id = identifier.add(group + "_ca")
        sa = SubordinateAuthority(
            resource_name=sub_id,
            subject=acmpca_native.CertificateAuthoritySubjectArgs(
                **{**base_subject, **dict(organizational_unit=group)}
            ),
            opts=depends_on(ca, certificate_l1, root_activation),
        )
        cert_id = sub_id.add("cert")
        cert = CertificateL2(resource_name=cert_id, ca=ca, certificate=sa, opts=depends_on(sa))
        level2_activation = Activation(
            resource_name=cert_id.add("validation"),
            cert=cert,
            ca=sa,
            certificate_chain=root_activation.complete_certificate_chain,
            opts=depends_on(sa, cert),
        )
        certs[f"root/{group}"] = level2_activation.complete_certificate_chain
        DefaultPermissions(
            resource_name=cert_id.add("permission"), cert=sa, opts=depends_on(sa, cert, level2_activation)
        )

        for app in apps:
            app_id = sub_id.add(app + "_ca")
            aa = SubordinateAuthority(
                resource_name=app_id,
                subject=acmpca_native.CertificateAuthoritySubjectArgs(
                    **{**base_subject, **dict(organizational_unit=group, common_name=app)}
                ),
                opts=depends_on(sa, cert, level2_activation),
            )
            cert_id = app_id.add("cert")
            cert = CertificateL3(resource_name=cert_id, ca=ca, certificate=aa, opts=depends_on(aa))
            level3_activation = Activation(
                resource_name=cert_id.add("validation"),
                cert=cert,
                ca=aa,
                certificate_chain=level2_activation.complete_certificate_chain,
                opts=depends_on(aa, cert),
            )
            certs[f"root/{group}/{app}"] = level3_activation.complete_certificate_chain

            DefaultPermissions(
                resource_name=cert_id.add("permission"), cert=aa, opts=depends_on(aa, cert, level3_activation)
            )

    return certs
