from typing import Optional, Sequence, Mapping

import pulumi
from pulumi_aws import alb, wafv2, get_caller_identity, iam
from pulumi_aws.ssm import get_parameter

from cloudlib.modules.acm.certificate import CertificateFactory
from cloudlib.modules.resource import BaseResourceMixin, BaseFactory
from cloudlib.modules.route53 import ZoneFactory
from cloudlib.modules.s3 import BucketPolicyFactory
from cloudlib.utils.names import IdentifierV2


class ALBBucketAccessLogPolicyFactory(BucketPolicyFactory):
    def __init__(self, accounts=None):
        self.accounts = accounts
        super().__init__()

    def create(self, bucket):
        identity = get_caller_identity()
        accounts = self.accounts or [identity.account_id]

        statements = []

        for account_id in accounts:
            statements.append(
                iam.GetPolicyDocumentStatementArgs(
                    sid=f"PolicyForALB_{account_id}",
                    effect="Allow",
                    principals=[
                        # account id below is and account id owning ALB (https://docs.aws.amazon.com/elasticloadbalancing/latest/application/enable-access-logging.html#access-log-create-bucket)
                        iam.GetPolicyDocumentStatementPrincipalArgs(
                            type="AWS", identifiers=[f"arn:aws:iam::127311923021:root"]
                        )
                    ],
                    actions=["s3:PutObject"],
                    resources=[f"arn:aws:s3:::{bucket}/*"],
                )
            )

        return super().create(bucket, extra_statements=statements)


class LoadBalancer(alb.LoadBalancer, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        access_logs: Optional[pulumi.Input[pulumi.InputType["LoadBalancerAccessLogsArgs"]]] = None,
        customer_owned_ipv4_pool: Optional[pulumi.Input[str]] = None,
        desync_mitigation_mode: Optional[pulumi.Input[str]] = None,
        drop_invalid_header_fields: Optional[pulumi.Input[bool]] = None,
        enable_cross_zone_load_balancing: Optional[pulumi.Input[bool]] = None,
        enable_deletion_protection: Optional[pulumi.Input[bool]] = None,
        enable_http2: Optional[pulumi.Input[bool]] = None,
        enable_tls_version_and_cipher_suite_headers: Optional[pulumi.Input[bool]] = None,
        enable_waf_fail_open: Optional[pulumi.Input[bool]] = None,
        enable_xff_client_port: Optional[pulumi.Input[bool]] = None,
        idle_timeout: Optional[pulumi.Input[int]] = None,
        internal: Optional[pulumi.Input[bool]] = None,
        ip_address_type: Optional[pulumi.Input[str]] = None,
        load_balancer_type: Optional[pulumi.Input[str]] = None,
        name_prefix: Optional[pulumi.Input[str]] = None,
        preserve_host_header: Optional[pulumi.Input[bool]] = None,
        security_groups: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        subnet_mappings: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["LoadBalancerSubnetMappingArgs"]]]]
        ] = None,
        subnets: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        xff_header_processing_mode: Optional[pulumi.Input[str]] = None,
        __props__=None,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=alb.LoadBalancer,
            resource_name=name,
            name=name,
            opts=opts,
            access_logs=access_logs,
            customer_owned_ipv4_pool=customer_owned_ipv4_pool,
            desync_mitigation_mode=desync_mitigation_mode,
            drop_invalid_header_fields=drop_invalid_header_fields,
            enable_cross_zone_load_balancing=enable_cross_zone_load_balancing,
            enable_deletion_protection=enable_deletion_protection,
            enable_http2=enable_http2,
            enable_tls_version_and_cipher_suite_headers=enable_tls_version_and_cipher_suite_headers,
            enable_waf_fail_open=enable_waf_fail_open,
            enable_xff_client_port=enable_xff_client_port,
            idle_timeout=idle_timeout,
            internal=internal,
            ip_address_type=ip_address_type,
            load_balancer_type=load_balancer_type,
            name_prefix=name_prefix,
            preserve_host_header=preserve_host_header,
            security_groups=security_groups,
            subnet_mappings=subnet_mappings,
            subnets=subnets,
            tags=tags,
            xff_header_processing_mode=xff_header_processing_mode,
            __props__=__props__,
        )


class ListenerRuleFactory(BaseFactory):
    def __init__(
        self,
        resource_name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        actions: Optional[pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ListenerRuleActionArgs"]]]]] = None,
        conditions: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ListenerRuleConditionArgs"]]]]
        ] = None,
        listener_arn: Optional[pulumi.Input[str]] = None,
        priority: Optional[pulumi.Input[int]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        __props__=None,
    ):
        super(ListenerRuleFactory, self).__init__(
            resource_name=resource_name,
            opts=opts,
            actions=actions,
            conditions=conditions,
            listener_arn=listener_arn,
            priority=priority,
            tags=tags,
        )

    def create(self):
        self._create(resource_type=alb.ListenerRule)


class ListenerFactory(BaseFactory):
    def __init__(
        self,
        resource_name: str,
        resource_id=None,
        opts: Optional[pulumi.ResourceOptions] = None,
        alpn_policy: Optional[pulumi.Input[str]] = None,
        certificate_arn: Optional[pulumi.Input[str]] = None,
        default_actions: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ListenerDefaultActionArgs"]]]]
        ] = None,
        load_balancer_arn: Optional[pulumi.Input[str]] = None,
        port: Optional[pulumi.Input[int]] = None,
        protocol: Optional[pulumi.Input[str]] = None,
        ssl_policy: Optional[pulumi.Input[str]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        __props__=None,
    ):
        if resource_id:
            self.arn = resource_id
        else:
            super().__init__(
                resource_name=resource_name,
                opts=opts,
                alpn_policy=alpn_policy,
                certificate_arn=certificate_arn,
                default_actions=default_actions,
                load_balancer_arn=load_balancer_arn,
                port=port,
                protocol=protocol,
                ssl_policy=ssl_policy,
                tags=tags,
                __props__=__props__,
            )

    def create(self):
        self.set_default(
            "default_actions",
            [
                alb.ListenerDefaultActionArgs(
                    type="fixed-response",
                    fixed_response=alb.ListenerDefaultActionFixedResponseArgs(
                        status_code=400, message_body="Invalid request", content_type="text/html"
                    ),
                )
            ],
        )
        return self._create(resource_type=alb.Listener)


class LoadBalancerFactory(BaseFactory):
    def __init__(
        self,
        resource_name: str,
        resource_id=None,
        opts: Optional[pulumi.ResourceOptions] = None,
        access_logs: Optional[pulumi.Input[pulumi.InputType["LoadBalancerAccessLogsArgs"]]] = None,
        customer_owned_ipv4_pool: Optional[pulumi.Input[str]] = None,
        desync_mitigation_mode: Optional[pulumi.Input[str]] = None,
        drop_invalid_header_fields: Optional[pulumi.Input[bool]] = None,
        enable_cross_zone_load_balancing: Optional[pulumi.Input[bool]] = None,
        enable_deletion_protection: Optional[pulumi.Input[bool]] = None,
        enable_http2: Optional[pulumi.Input[bool]] = None,
        enable_tls_version_and_cipher_suite_headers: Optional[pulumi.Input[bool]] = None,
        enable_waf_fail_open: Optional[pulumi.Input[bool]] = None,
        enable_xff_client_port: Optional[pulumi.Input[bool]] = None,
        idle_timeout: Optional[pulumi.Input[int]] = None,
        internal: Optional[pulumi.Input[bool]] = None,
        ip_address_type: Optional[pulumi.Input[str]] = None,
        load_balancer_type: Optional[pulumi.Input[str]] = None,
        name_prefix: Optional[pulumi.Input[str]] = None,
        preserve_host_header: Optional[pulumi.Input[bool]] = None,
        security_groups: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        subnet_mappings: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["LoadBalancerSubnetMappingArgs"]]]]
        ] = None,
        subnets: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        xff_header_processing_mode: Optional[pulumi.Input[str]] = None,
        resources=None,
        __props__=None,
    ):
        self.listeners = []
        if resource_id:
            self.arn = resource_id
            self.resources = resources
        else:
            super().__init__(
                resource_name=resource_name,
                name=IdentifierV2(resource_name).spinal,
                opts=opts,
                access_logs=access_logs,
                customer_owned_ipv4_pool=customer_owned_ipv4_pool,
                desync_mitigation_mode=desync_mitigation_mode,
                drop_invalid_header_fields=drop_invalid_header_fields,
                enable_cross_zone_load_balancing=enable_cross_zone_load_balancing,
                enable_deletion_protection=enable_deletion_protection,
                enable_http2=enable_http2,
                enable_tls_version_and_cipher_suite_headers=enable_tls_version_and_cipher_suite_headers,
                enable_waf_fail_open=enable_waf_fail_open,
                enable_xff_client_port=enable_xff_client_port,
                idle_timeout=idle_timeout,
                internal=internal,
                ip_address_type=ip_address_type,
                load_balancer_type=load_balancer_type,
                name_prefix=name_prefix,
                preserve_host_header=preserve_host_header,
                security_groups=security_groups,
                subnet_mappings=subnet_mappings,
                subnets=subnets,
                tags=tags,
                xff_header_processing_mode=xff_header_processing_mode,
                __props__=__props__,
            )
        self.waf = None

    def with_https(self, domain, zone_id, subject_alternative_names=None, validator=None, **kwargs):
        cert_factory = CertificateFactory(
            resource_name=self.resource_name,
            domain_name=domain,
            subject_alternative_names=subject_alternative_names,
            **kwargs,
        )
        certificate = cert_factory.create()

        if validator:
            validator(certificate)
        else:
            zone_factory = ZoneFactory(self.resource_name, zone_id)
            zone_factory.validate_certificate(certificate)

        kwargs.setdefault("protocol", "HTTPS")
        kwargs.setdefault("ssl_policy", "ELBSecurityPolicy-TLS13-1-2-2021-06")

        self.listeners.append(
            ListenerFactory(self.identifier.add("https"), certificate_arn=certificate.arn, port=443, **kwargs)
        )

        return certificate

    def with_waf(self, arn=True):
        self.waf = arn

    def get_https_listener(self) -> alb.AwaitableGetListenerResult:
        return self.arn.apply(lambda arn: alb.get_listener(load_balancer_arn=arn, port=443))

    def get(self):
        if self.arn:
            return alb.get_load_balancer(arn=self.arn)
        else:
            return self.create()

    def create_https_rule(
        self,
        name,
        target_group=None,
        domain_name=None,
        opts: Optional[pulumi.ResourceOptions] = None,
        actions: Optional[pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ListenerRuleActionArgs"]]]]] = None,
        conditions: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ListenerRuleConditionArgs"]]]]
        ] = None,
        listener_arn: Optional[pulumi.Input[str]] = None,
        priority: Optional[pulumi.Input[int]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        __props__=None,
    ) -> ListenerRuleFactory:
        if not actions:
            actions = [alb.ListenerRuleActionArgs(type="forward", target_group_arn=target_group.arn)]

        if not conditions:
            conditions = [
                alb.ListenerRuleConditionArgs(host_header=alb.ListenerRuleConditionHostHeaderArgs(values=[domain_name]))
            ]

        return ListenerRuleFactory(
            resource_name=name,
            opts=opts,
            actions=actions,
            conditions=conditions,
            listener_arn=listener_arn or self.resources.get("lb_shared_listener_443") or self.get_https_listener(),
            priority=priority,
            tags=tags,
        )

    def create(self, with_exports=True) -> alb.LoadBalancer:
        self.set_default("enable_deletion_protection", True)
        self.set_default("drop_invalid_header_fields", True)

        lb: alb.LoadBalancer = self._create(resource_type=alb.LoadBalancer)

        if self.waf:
            if isinstance(self.waf, bool):
                waf_arn = get_parameter(name="/security/waf/regional/arn").value
            else:
                waf_arn = self.waf
            wafv2.WebAclAssociation(IdentifierV2("lb_" + self.name).snake, resource_arn=lb.arn, web_acl_arn=waf_arn)

        for listener_factory in self.listeners:
            listener_factory.kwargs["load_balancer_arn"] = lb.arn
            listener = listener_factory.create()
            if with_exports:
                pulumi.export(f"lb_shared_listener_{listener_factory.kwargs['port']}", listener.arn)

        if with_exports:
            pulumi.export("lb_shared_arn", lb.arn)
            pulumi.export("lb_shared_zone_id", lb.zone_id)
            pulumi.export("lb_shared_dns_name", lb.dns_name)

        return lb

    def with_logging(self, bucket):
        self.set_default("access_logs", {"bucket": bucket, "enabled": True})


class TargetGroupFactory(BaseFactory):
    def __init__(
        self,
        resource_name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        connection_termination: Optional[pulumi.Input[bool]] = None,
        deregistration_delay: Optional[pulumi.Input[int]] = None,
        health_check: Optional[pulumi.Input[pulumi.InputType["TargetGroupHealthCheckArgs"]]] = None,
        ip_address_type: Optional[pulumi.Input[str]] = None,
        lambda_multi_value_headers_enabled: Optional[pulumi.Input[bool]] = None,
        load_balancing_algorithm_type: Optional[pulumi.Input[str]] = None,
        load_balancing_cross_zone_enabled: Optional[pulumi.Input[str]] = None,
        name_prefix: Optional[pulumi.Input[str]] = None,
        port: Optional[pulumi.Input[int]] = None,
        preserve_client_ip: Optional[pulumi.Input[str]] = None,
        protocol: Optional[pulumi.Input[str]] = None,
        protocol_version: Optional[pulumi.Input[str]] = None,
        proxy_protocol_v2: Optional[pulumi.Input[bool]] = None,
        slow_start: Optional[pulumi.Input[int]] = None,
        stickiness: Optional[pulumi.Input[pulumi.InputType["TargetGroupStickinessArgs"]]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        target_failovers: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["TargetGroupTargetFailoverArgs"]]]]
        ] = None,
        target_type: Optional[pulumi.Input[str]] = None,
        vpc_id: Optional[pulumi.Input[str]] = None,
        __props__=None,
    ):
        super().__init__(
            resource_name=resource_name,
            opts=opts,
            connection_termination=connection_termination,
            deregistration_delay=deregistration_delay,
            health_check=health_check,
            ip_address_type=ip_address_type,
            lambda_multi_value_headers_enabled=lambda_multi_value_headers_enabled,
            load_balancing_algorithm_type=load_balancing_algorithm_type,
            load_balancing_cross_zone_enabled=load_balancing_cross_zone_enabled,
            name_prefix=name_prefix,
            name=IdentifierV2(resource_name).spinal,
            port=port,
            preserve_client_ip=preserve_client_ip,
            protocol=protocol,
            protocol_version=protocol_version,
            proxy_protocol_v2=proxy_protocol_v2,
            slow_start=slow_start,
            stickiness=stickiness,
            tags=tags,
            target_failovers=target_failovers,
            target_type=target_type,
            vpc_id=vpc_id,
        )

    def create(self) -> alb.TargetGroup:
        return self._create(resource_type=alb.TargetGroup)
