from os.path import dirname
from typing import Optional, Sequence, Mapping
from collections import OrderedDict

import pulumi
from pulumi_aws import cfg, iam as piam, get_region, get_caller_identity

from cloudlib.modules import iam
from cloudlib.modules.resource import BaseFactory
from cloudlib.modules.s3 import BucketFactory, BucketPolicyFactory
from cloudlib.utils.names import IdentifierV2


class ConfigS3PolicyFactory(BucketPolicyFactory):
    def __init__(self, bucket_name, accounts):
        self.bucket_name = bucket_name
        self.accounts = accounts
        super().__init__()

    def create(self, bucket):
        statements = []

        for account in self.accounts:
            statements.append(
                piam.GetPolicyDocumentStatementArgs(
                    sid=f"AWSConfigBucketPermissionsCheck_{account}",
                    effect="Allow",
                    principals=[
                        piam.GetPolicyDocumentStatementPrincipalArgs(
                            type="Service",
                            identifiers=["config.amazonaws.com"],
                        )
                    ],
                    actions=["s3:GetBucketAcl", "s3:ListBucket"],
                    resources=[f"arn:aws:s3:::{bucket}"],
                    conditions=[
                        piam.GetPolicyDocumentStatementConditionArgs(
                            test="StringEquals", variable="AWS:SourceAccount", values=[account]
                        )
                    ],
                )
            )

            statements.append(
                piam.GetPolicyDocumentStatementArgs(
                    sid=f"AWSConfigBucketDelivery_{account}",
                    effect="Allow",
                    principals=[
                        piam.GetPolicyDocumentStatementPrincipalArgs(
                            type="Service",
                            identifiers=["config.amazonaws.com"],
                        )
                    ],
                    actions=["s3:PutObject"],
                    resources=[
                        f"arn:aws:s3:::{bucket}/AWSLogs/{account}/Config/*",
                    ],
                    conditions=[
                        piam.GetPolicyDocumentStatementConditionArgs(
                            test="StringEquals", variable="s3:x-amz-acl", values=["bucket-owner-full-control"]
                        ),
                        piam.GetPolicyDocumentStatementConditionArgs(
                            test="StringEquals", variable="AWS:SourceAccount", values=[account]
                        ),
                    ],
                ),
            )

        return super().create(bucket, extra_statements=statements)


class RecorderRole(iam.Role):
    def __init__(
        self,
        name,
        opts: Optional[pulumi.ResourceOptions] = None,
        description: Optional[pulumi.Input[str]] = None,
        force_detach_policies: Optional[pulumi.Input[bool]] = None,
        inline_policies: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["RoleInlinePolicyArgs"]]]]
        ] = None,
        # managed_policy_arns: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        max_session_duration: Optional[pulumi.Input[int]] = None,
        name_prefix: Optional[pulumi.Input[str]] = None,
        path: Optional[pulumi.Input[str]] = None,
        permissions_boundary: Optional[pulumi.Input[str]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        __props__=None,
    ):
        assume_role_policy = piam.get_policy_document(
            statements=[
                piam.GetPolicyDocumentStatementArgs(
                    effect="Allow",
                    principals=[
                        piam.GetPolicyDocumentStatementPrincipalArgs(
                            type="Service",
                            identifiers=["config.amazonaws.com"],
                        )
                    ],
                    actions=["sts:AssumeRole"],
                )
            ]
        )

        managed_policy_arns = ["arn:aws:iam::aws:policy/service-role/AWS_ConfigRole"]

        super().__init__(
            name,
            opts,
            assume_role_policy.json,
            description,
            force_detach_policies,
            inline_policies,
            managed_policy_arns,
            max_session_duration,
            name_prefix,
            path,
            permissions_boundary,
            tags,
            __props__,
        )


class RecorderFactory(BaseFactory):
    def __init__(
        self,
        config,
        opts: Optional[pulumi.ResourceOptions] = None,
        name: Optional[pulumi.Input[str]] = None,
        recording_group: Optional[pulumi.Input[pulumi.InputType["RecorderRecordingGroupArgs"]]] = None,
        role_arn: Optional[pulumi.Input[str]] = None,
    ):
        super().__init__(
            resource_name=name,
            opts=opts,
            name=name,
            recording_group=recording_group,
            role_arn=role_arn,
        )
        self.delivery_factory = None
        self.bucket_factory = None
        self.packs = OrderedDict(
            config.get_object(
                "aws-conformance-packs",
                {
                    "HIPAA-Security": "Operational-Best-Practices-for-HIPAA-Security",
                    "AWS-IAM": "Operational-Best-Practices-for-AWS-Identity-and-Access-Management",
                    "Lambda-Best-Practices": "Security-Best-Practices-for-Lambda",
                },
            )
        )

        default_config = {
            "Lambda-Best-Practices": {
                "LambdaFunctionSettingsCheckParamRuntime": "nodejs18.x, nodejs16.x, nodejs14.x, python3.11, python3.10, python3.9, python3.8, python3.7, python3.6"
            },
            "HIPAA-Security": {},
        }

        default_config.update(config.get_object("aws-conformance-packs-parameters", {}))

        self.packs_config = default_config

    def with_s3_delivery(self, name, **kwargs):
        opts = self.kwargs.get("opts")
        kwargs.setdefault("opts", opts)

        identity = get_caller_identity(opts=opts)
        region = get_region(opts=opts)

        self.bucket_factory = BucketFactory(name=name, **kwargs)

        self.bucket_factory.with_policy(
            policy=ConfigS3PolicyFactory(bucket_name=self.bucket_factory.full_name, accounts=[identity.account_id])
        )
        self.delivery_factory = DeliveryChannelFactory(
            name=name + "-delivery", s3_bucket_name=self.bucket_factory.ref("bucket"), opts=self.kwargs.get("opts")
        )

        return self.bucket_factory

    def with_pack(self, name, pack):
        self.packs[name] = pack

    def create(self):
        role = self.kwargs.get("role_arn")
        if not role:
            role = RecorderRole(IdentifierV2(self.kwargs.get("name") + "Role").camel, opts=self.kwargs.get("opts")).arn

        self.setdefault("role_arn", role)
        if self.bucket_factory:
            bucket = self.bucket_factory.create()
            pulumi.export("config_bucket", bucket.bucket)

        recorder = self._create(cfg.Recorder)
        base_dir = dirname(__file__)
        for name, pack in self.packs.items():
            with open(f"{base_dir}/../../rules/aws-config-conformance-packs/{pack}.yaml") as pack_body:
                cfg.ConformancePack(
                    IdentifierV2(f"pack_{pack}").snake,
                    name=name,
                    input_parameters=[
                        dict(parameter_name=name, parameter_value=value)
                        for name, value in self.packs_config.get(name, {}).items()
                    ],
                    template_body=pack_body.read(),
                    opts=pulumi.ResourceOptions(depends_on=[recorder]),
                )

        if self.delivery_factory:
            channel = self.delivery_factory.create()

            cfg.RecorderStatus(
                self.kwargs.get("name") + "-status",
                is_enabled=True,
                name=self.kwargs.get("name"),
                opts=pulumi.ResourceOptions(depends_on=[channel]),
            )

        return recorder


class DeliveryChannelFactory(BaseFactory):
    def __init__(
        self,
        opts: Optional[pulumi.ResourceOptions] = None,
        name: Optional[pulumi.Input[str]] = None,
        s3_bucket_name: Optional[pulumi.Input[str]] = None,
        s3_key_prefix: Optional[pulumi.Input[str]] = None,
        s3_kms_key_arn: Optional[pulumi.Input[str]] = None,
        snapshot_delivery_properties: Optional[
            pulumi.Input[pulumi.InputType["DeliveryChannelSnapshotDeliveryPropertiesArgs"]]
        ] = None,
        sns_topic_arn: Optional[pulumi.Input[str]] = None,
    ):
        super().__init__(
            resource_name=name,
            opts=opts,
            name=name,
            s3_bucket_name=s3_bucket_name,
            s3_key_prefix=s3_key_prefix,
            s3_kms_key_arn=s3_kms_key_arn,
            snapshot_delivery_properties=snapshot_delivery_properties,
            sns_topic_arn=sns_topic_arn,
        )

    def create(self):
        channel = self._create(cfg.DeliveryChannel)

        return channel
