import json
from typing import Optional, Sequence, Mapping, overload
import pulumi
from pulumi_aws.cloudtrail import TrailArgs

from cloudlib.modules.cloudwatch import LogGroupFactory
from cloudlib.modules.iam import RoleFactory, Policy, Statement
from cloudlib.modules.kms import KeyFactory
from cloudlib.modules.resource import BaseFactory
from pulumi_aws import cloudtrail, iam, get_caller_identity, get_region, s3

from cloudlib.modules.s3 import BucketFactory, BucketPolicyFactory, S3BucketAccessLogPolicyFactory
from cloudlib.utils.names import IdentifierV2

current_caller_identity = get_caller_identity()
region = get_region()


class TrailS3PolicyFactory(BucketPolicyFactory):
    def __init__(self, trail_name, accounts=None):
        self.trail_name = trail_name
        self.accounts = accounts
        super().__init__()

    def create(self, bucket):
        arn = f"arn:aws:s3:::{bucket}"

        if not self.accounts:
            self.accounts = [current_caller_identity.account_id]

        return super().create(
            bucket,
            extra_statements=[
                iam.GetPolicyDocumentStatementArgs(
                    sid="AWSCloudTrailAclCheck",
                    effect="Allow",
                    principals=[
                        iam.GetPolicyDocumentStatementPrincipalArgs(
                            type="Service",
                            identifiers=["cloudtrail.amazonaws.com"],
                        )
                    ],
                    actions=["s3:GetBucketAcl"],
                    resources=[arn],
                    conditions=[
                        iam.GetPolicyDocumentStatementConditionArgs(
                            test="StringEquals",
                            variable="AWS:SourceArn",
                            values=[
                                f"arn:aws:cloudtrail:{region.id}:{account_id}:trail/{self.trail_name}"
                                for account_id in self.accounts
                            ],
                        )
                    ],
                ),
            ]
            + [
                iam.GetPolicyDocumentStatementArgs(
                    sid=f"AWSCloudTrailWrite_{account_id}",
                    effect="Allow",
                    principals=[
                        iam.GetPolicyDocumentStatementPrincipalArgs(
                            type="Service",
                            identifiers=["cloudtrail.amazonaws.com"],
                        )
                    ],
                    actions=["s3:PutObject"],
                    resources=[f"{arn}/AWSLogs/{account_id}/*"],
                    conditions=[
                        iam.GetPolicyDocumentStatementConditionArgs(
                            test="StringEquals",
                            variable="s3:x-amz-acl",
                            values=["bucket-owner-full-control"],
                        ),
                        iam.GetPolicyDocumentStatementConditionArgs(
                            test="StringEquals",
                            variable="AWS:SourceArn",
                            values=[f"arn:aws:cloudtrail:{region.id}:{account_id}:trail/{self.trail_name}"],
                        ),
                    ],
                )
                for account_id in self.accounts
            ],
        )


class TrailFactory(BaseFactory):
    def __init__(
        self,
        opts: Optional[pulumi.ResourceOptions] = None,
        advanced_event_selectors: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["TrailAdvancedEventSelectorArgs"]]]]
        ] = None,
        cloud_watch_logs_group_arn: Optional[pulumi.Input[str]] = None,
        cloud_watch_logs_role_arn: Optional[pulumi.Input[str]] = None,
        enable_log_file_validation: Optional[pulumi.Input[bool]] = None,
        enable_logging: Optional[pulumi.Input[bool]] = None,
        event_selectors: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["TrailEventSelectorArgs"]]]]
        ] = None,
        include_global_service_events: Optional[pulumi.Input[bool]] = None,
        insight_selectors: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["TrailInsightSelectorArgs"]]]]
        ] = None,
        is_multi_region_trail: Optional[pulumi.Input[bool]] = None,
        is_organization_trail: Optional[pulumi.Input[bool]] = None,
        kms_key_id: Optional[pulumi.Input[str]] = None,
        name: Optional[pulumi.Input[str]] = None,
        s3_bucket_name: Optional[pulumi.Input[str]] = None,
        s3_key_prefix: Optional[pulumi.Input[str]] = None,
        sns_topic_name: Optional[pulumi.Input[str]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        bucket_factory=None,
        accounts=None,
    ):
        super().__init__(
            resource_name=name,
            opts=opts,
            advanced_event_selectors=advanced_event_selectors,
            cloud_watch_logs_group_arn=cloud_watch_logs_group_arn,
            cloud_watch_logs_role_arn=cloud_watch_logs_role_arn,
            enable_log_file_validation=enable_log_file_validation,
            enable_logging=enable_logging,
            event_selectors=event_selectors,
            include_global_service_events=include_global_service_events,
            insight_selectors=insight_selectors,
            is_multi_region_trail=is_multi_region_trail,
            is_organization_trail=is_organization_trail,
            kms_key_id=kms_key_id,
            name=name,
            s3_bucket_name=s3_bucket_name,
            s3_key_prefix=s3_key_prefix,
            sns_topic_name=sns_topic_name,
            tags=tags,
        )
        self.delivery_factory = None
        self.bucket_factory = bucket_factory
        self.log_group_factory = None
        self.key_factory = None
        self.accounts = accounts or [current_caller_identity.account_id]

    def with_cloudwatch(self, name):
        self.log_group_factory = LogGroupFactory(name)

    def create(self) -> cloudtrail.Trail:
        if not self.kwargs.get("s3_bucket_name"):
            if not self.bucket_factory:
                self.bucket_factory = BucketFactory("aws-cloudtrail")
                self.bucket_factory.with_policy(TrailS3PolicyFactory(self.kwargs.get("name")))

            if not self.bucket_factory.kwargs.get("loggings"):
                cloudtrail_log_factory = BucketFactory("aws-cloudtrail-log")
                cloudtrail_log_factory.with_policy(S3BucketAccessLogPolicyFactory())

                cloudtrail_log = cloudtrail_log_factory.create()

                self.bucket_factory.setdefault("loggings", [s3.BucketLoggingArgs(target_bucket=cloudtrail_log.bucket)])

            bucket = self.bucket_factory.create()
            self.setdefault("s3_bucket_name", bucket.bucket)

        self.setdefault("include_global_service_events", False)
        self.setdefault("enable_log_file_validation", True)
        self.setdefault("include_global_service_events", True)
        self.setdefault("is_multi_region_trail", True)

        if not self.kwargs.get("kms_key_id"):
            self.key_factory = KeyFactory("cloudtrail-cloud-audit-trail")
            self.key_factory.with_alias("cloudtrail-cloud-audit-trail")
            self.key_factory.allow_usage(self)

            key = self.key_factory.create()
            self.setdefault("kms_key_id", key.arn)

        self.setdefault(
            "event_selectors",
            [
                cloudtrail.TrailEventSelectorArgs(
                    data_resources=[
                        cloudtrail.TrailEventSelectorDataResourceArgs(
                            type="AWS::DynamoDB::Table", values=["arn:aws:dynamodb"]
                        ),
                        cloudtrail.TrailEventSelectorDataResourceArgs(
                            type="AWS::Lambda::Function", values=["arn:aws:lambda"]
                        ),
                        cloudtrail.TrailEventSelectorDataResourceArgs(type="AWS::S3::Object", values=["arn:aws:s3"]),
                    ],
                    exclude_management_event_sources=["kms.amazonaws.com"],
                    include_management_events=True,
                    read_write_type="All",
                )
            ],
        )

        if self.log_group_factory:
            log_group = self.log_group_factory.create()
            role = self.cloudtrail_cloudwatch_log_role(arn=self.log_group_factory.arn() + ":*")

            self.setdefault("cloud_watch_logs_group_arn", log_group.arn.apply(lambda arn: f"{arn}:*"))
            self.setdefault("cloud_watch_logs_role_arn", role.arn)

        return self._create(cloudtrail.Trail)

    def cloudtrail_cloudwatch_log_role(self, arn):
        role_factory = RoleFactory(
            "CloudTrail" + IdentifierV2(self.kwargs.get("name")).camel + "LogRole",
            assume_role_policy=json.dumps(
                {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Sid": "",
                            "Effect": "Allow",
                            "Principal": {"Service": "cloudtrail.amazonaws.com"},
                            "Action": "sts:AssumeRole",
                        }
                    ],
                }
            ),
        )

        role_factory.with_inline_policy(
            Policy(
                "allow_cloudwatch_logs",
                [
                    Statement(
                        effect="Allow",
                        actions=["logs:CreateLogStream", "logs:PutLogEvents"],
                        resources=arn,
                    ),
                ],
            )
        )
        return role_factory.create()

    def kms_policy(self, suffix=""):
        return [
            {
                "Sid": f"Allow CloudTrail {account_id}",
                "Effect": "Allow",
                "Principal": {"Service": self.service()},
                "Action": "kms:GenerateDataKey*",
                "Resource": "*",
                "Condition": {
                    "StringEquals": {
                        "aws:SourceArn": f"arn:aws:cloudtrail:{region.id}:{account_id}:trail/{self.kwargs.get('name')}"
                    },
                    "StringLike": {
                        "kms:EncryptionContext:aws:cloudtrail:arn": f"arn:aws:cloudtrail:*:{account_id}:trail/*"
                    },
                },
            }
            for account_id in self.accounts
        ]

    def service(self):
        return "cloudtrail.amazonaws.com"

    def arn(self):
        return super().arn(resource_type="cloudtrail", resource_id="trail/" + self.kwargs.get("name"))
