from typing import Optional, Mapping, Sequence

import pulumi

from cloudlib.modules.iam import RoleFactory
from cloudlib.modules.kms import KeyFactory
from cloudlib.modules.resource import BaseFactory
from pulumi_aws import cloudwatch

from cloudlib.utils.names import IdentifierV2


class LogGroupFactory(BaseFactory):
    def __init__(
        self,
        name,
        opts: Optional[pulumi.ResourceOptions] = None,
        kms_key_id: Optional[pulumi.Input[str]] = None,
        name_prefix: Optional[pulumi.Input[str]] = None,
        retention_in_days: Optional[pulumi.Input[int]] = None,
        skip_destroy: Optional[pulumi.Input[bool]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
    ):
        if not name.startswith("/"):
            name = "/" + name

        super().__init__(
            resource_name=name,
            opts=opts,
            kms_key_id=kms_key_id,
            name=name,
            name_prefix=name_prefix,
            retention_in_days=retention_in_days,
            skip_destroy=skip_destroy,
            tags=tags,
        )
        self.key_factory = None
        self.role_factory: RoleFactory = None

    def with_owner_role(self):
        self.role_factory = RoleFactory("LogGroup" + IdentifierV2(self.kwargs.get("name")).camel + "Owner")

    def create(self):
        self.setdefault("retention_in_days", 365)
        self.setdefault("skip_destroy", True)

        if not self.kwargs.get("kms_key_id"):
            self.key_factory = KeyFactory(self.kwargs.get("name"))
            self.key_factory.with_alias(self.kwargs.get("name") + "-loggroup")
            self.key_factory.allow_usage(self)
            self.key_factory.create()

            self.setdefault("kms_key_id", self.key_factory.ref("arn"))

        return self._create(cloudwatch.LogGroup)

    def service(self):
        return "logs.us-east-1.amazonaws.com"

    def arn(self):
        return super().arn(resource_type="logs", resource_id="log-group:" + self.kwargs.get("name"))


class MetricAlarmFactory(BaseFactory):
    def __init__(
        self,
        name: Optional[pulumi.Input[str]],
        opts: Optional[pulumi.ResourceOptions] = None,
        actions_enabled: Optional[pulumi.Input[bool]] = None,
        alarm_actions: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        alarm_description: Optional[pulumi.Input[str]] = None,
        comparison_operator: Optional[pulumi.Input[str]] = None,
        datapoints_to_alarm: Optional[pulumi.Input[int]] = None,
        dimensions: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        evaluate_low_sample_count_percentiles: Optional[pulumi.Input[str]] = None,
        evaluation_periods: Optional[pulumi.Input[int]] = None,
        extended_statistic: Optional[pulumi.Input[str]] = None,
        insufficient_data_actions: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        metric_name: Optional[pulumi.Input[str]] = None,
        metric_queries: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["MetricAlarmMetricQueryArgs"]]]]
        ] = None,
        namespace: Optional[pulumi.Input[str]] = None,
        ok_actions: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        period: Optional[pulumi.Input[int]] = None,
        statistic: Optional[pulumi.Input[str]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        threshold: Optional[pulumi.Input[float]] = None,
        threshold_metric_id: Optional[pulumi.Input[str]] = None,
        treat_missing_data: Optional[pulumi.Input[str]] = None,
        unit: Optional[pulumi.Input[str]] = None,
    ):
        super().__init__(
            resource_name=name,
            opts=opts,
            actions_enabled=actions_enabled,
            alarm_actions=alarm_actions,
            alarm_description=alarm_description,
            comparison_operator=comparison_operator,
            datapoints_to_alarm=datapoints_to_alarm,
            dimensions=dimensions,
            evaluate_low_sample_count_percentiles=evaluate_low_sample_count_percentiles,
            evaluation_periods=evaluation_periods,
            extended_statistic=extended_statistic,
            insufficient_data_actions=insufficient_data_actions,
            metric_name=metric_name,
            metric_queries=metric_queries,
            name=name,
            namespace=namespace,
            ok_actions=ok_actions,
            period=period,
            statistic=statistic,
            tags=tags,
            threshold=threshold,
            threshold_metric_id=threshold_metric_id,
            treat_missing_data=treat_missing_data,
            unit=unit,
        )

    def create(self):
        self.setdefault("evaluation_periods", 2)
        self.setdefault("statistic", "Maximum")
        self.setdefault("period", 60)
        self.setdefault("namespace", "AWS/Logs")
        self._create(cloudwatch.MetricAlarm)


class LogMetricFilterFactory(BaseFactory):
    def __init__(
        self,
        name: Optional[pulumi.Input[str]],
        pattern: Optional[pulumi.Input[str]],
        opts: Optional[pulumi.ResourceOptions] = None,
        log_group_name: Optional[pulumi.Input[str]] = None,
        metric_transformation: Optional[
            pulumi.Input[pulumi.InputType["LogMetricFilterMetricTransformationArgs"]]
        ] = None,
    ):
        super().__init__(
            resource_name=name,
            opts=opts,
            log_group_name=log_group_name,
            metric_transformation=metric_transformation,
            name=name,
            pattern=pattern,
        )
        self.alarm_factory = None
        self.metric_name = None
        self.filter_namespace = None

    def apply(self, name, namespace, value, default_value=None, dimensions=None, unit=None):
        self.metric_name = name
        self.filter_namespace = namespace
        self.setdefault(
            "metric_transformation",
            cloudwatch.LogMetricFilterMetricTransformationArgs(
                name=name,
                namespace=namespace,
                value=value,
                default_value=default_value,
                dimensions=dimensions,
                unit=unit,
            ),
        )

    def create(self):
        metric = self._create(cloudwatch.LogMetricFilter)
        if self.alarm_factory:
            self.alarm_factory.create()

        return metric

    def with_alarm(self, sns):
        self.alarm_factory = MetricAlarmFactory(
            name=self.name,
            threshold=1,
            comparison_operator="GreaterThanOrEqualToThreshold",
            metric_name=self.metric_name,
            namespace=self.filter_namespace,
            alarm_description=f"Alert about: {self.name}",
            alarm_actions=[sns],
            period=60,
            datapoints_to_alarm=1,
            evaluation_periods=1,
        )
