import json
from typing import Optional, Sequence, Mapping, overload

import pulumi
from pulumi_aws import ecs, get_region, cloudwatch, lb, iam
from pulumi_aws.iam import RoleInlinePolicyArgs

from cloudlib.modules.alb import LoadBalancerFactory, TargetGroupFactory
from cloudlib.modules.iam import Role
from cloudlib.modules.resource import BaseResourceMixin, BaseFactory
from cloudlib.modules.route53 import ZoneFactory
from cloudlib.utils.names import IdentifierV2


class Cluster(ecs.Cluster, BaseResourceMixin):
    def __init__(
        self,
        opts: Optional[pulumi.ResourceOptions] = None,
        configuration: Optional[pulumi.Input[pulumi.InputType["ClusterConfigurationArgs"]]] = None,
        name: Optional[pulumi.Input[str]] = None,
        service_connect_defaults: Optional[pulumi.Input[pulumi.InputType["ClusterServiceConnectDefaultsArgs"]]] = None,
        settings: Optional[pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ClusterSettingArgs"]]]]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        __props__=None,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=ecs.Cluster,
            resource_name=name,
            with_prefix=True,
            opts=opts,
            configuration=configuration,
            name=name,
            service_connect_defaults=service_connect_defaults,
            settings=settings,
            tags=tags,
            __props__=__props__,
        )

    def with_fargate(self):
        return ecs.ClusterCapacityProviders(
            self.identifier,
            cluster_name=self.name,
            capacity_providers=["FARGATE"],
            default_capacity_provider_strategies=[
                ecs.ClusterCapacityProvidersDefaultCapacityProviderStrategyArgs(
                    base=1,
                    weight=100,
                    capacity_provider="FARGATE",
                )
            ],
        )


class ContainerDefinition:
    def __init__(
        self,
        cpu,
        image,
        memory,
        name,
        volumes_from=None,
        entrypoint=None,
        command=None,
        port_mapping=None,
        mount_points=None,
        essential=True,
        healthcheck=None,
        environment=None,
        secrets=None,
        log_configuration=None,
    ):
        self.container = dict(
            cpu=cpu,
            image=image,
            memory=memory,
            name=name,
            volumesFrom=volumes_from,
            entrypoint=entrypoint,
            command=command,
            portMappings=port_mapping or [],
            mountPoints=mount_points,
            essential=essential,
            healthcheck=healthcheck,
            environment=environment or [],
            secrets=secrets or [],
            logConfiguration=(
                log_configuration
                or {
                    "logDriver": "awslogs",
                    "options": {
                        "awslogs-group": f"/ecs/{name}",
                        "awslogs-stream-prefix": "task",
                        "awslogs-region": get_region().id,
                    },
                }
            ),
        )
        self.outputs = {}

    def add_environment_variable(self, name, value):
        self.container["environment"].append({"name": name, "value": value})

    def add_secret(self, name, value):
        self.container["secrets"].append({"name": name, "valueFrom": value})

    def add_port_mapping(self, mapping):
        if isinstance(mapping, int):
            self.container["portMappings"].append(dict(containerPort=mapping, name=f"port_{mapping}"))
        else:
            self.container["portMappings"].append(mapping)


class ServiceFactory(BaseFactory):
    def __init__(
        self,
        resource_name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        alarms: Optional[pulumi.Input[pulumi.InputType["ServiceAlarmsArgs"]]] = None,
        capacity_provider_strategies: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ServiceCapacityProviderStrategyArgs"]]]]
        ] = None,
        cluster: Optional[pulumi.Input[str]] = None,
        deployment_circuit_breaker: Optional[
            pulumi.Input[pulumi.InputType["ServiceDeploymentCircuitBreakerArgs"]]
        ] = None,
        deployment_controller: Optional[pulumi.Input[pulumi.InputType["ServiceDeploymentControllerArgs"]]] = None,
        deployment_maximum_percent: Optional[pulumi.Input[int]] = None,
        deployment_minimum_healthy_percent: Optional[pulumi.Input[int]] = None,
        desired_count: Optional[pulumi.Input[int]] = None,
        enable_ecs_managed_tags: Optional[pulumi.Input[bool]] = None,
        enable_execute_command: Optional[pulumi.Input[bool]] = None,
        force_new_deployment: Optional[pulumi.Input[bool]] = None,
        health_check_grace_period_seconds: Optional[pulumi.Input[int]] = None,
        iam_role: Optional[pulumi.Input[str]] = None,
        launch_type: Optional[pulumi.Input[str]] = None,
        load_balancers: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ServiceLoadBalancerArgs"]]]]
        ] = None,
        network_configuration: Optional[pulumi.Input[pulumi.InputType["ServiceNetworkConfigurationArgs"]]] = None,
        ordered_placement_strategies: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ServiceOrderedPlacementStrategyArgs"]]]]
        ] = None,
        placement_constraints: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ServicePlacementConstraintArgs"]]]]
        ] = None,
        platform_version: Optional[pulumi.Input[str]] = None,
        propagate_tags: Optional[pulumi.Input[str]] = None,
        scheduling_strategy: Optional[pulumi.Input[str]] = None,
        service_connect_configuration: Optional[
            pulumi.Input[pulumi.InputType["ServiceServiceConnectConfigurationArgs"]]
        ] = None,
        service_registries: Optional[pulumi.Input[pulumi.InputType["ServiceServiceRegistriesArgs"]]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        task_definition: Optional[pulumi.Input[str]] = None,
        triggers: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        wait_for_steady_state: Optional[pulumi.Input[bool]] = None,
        __props__=None,
    ):
        super().__init__(
            resource_name=resource_name,
            name=IdentifierV2(resource_name).spinal,
            opts=opts,
            alarms=alarms,
            capacity_provider_strategies=capacity_provider_strategies,
            cluster=cluster,
            deployment_circuit_breaker=deployment_circuit_breaker,
            deployment_controller=deployment_controller,
            deployment_maximum_percent=deployment_maximum_percent,
            deployment_minimum_healthy_percent=deployment_minimum_healthy_percent,
            desired_count=desired_count,
            enable_ecs_managed_tags=enable_ecs_managed_tags,
            enable_execute_command=enable_execute_command,
            force_new_deployment=force_new_deployment,
            health_check_grace_period_seconds=health_check_grace_period_seconds,
            iam_role=iam_role,
            launch_type=launch_type,
            load_balancers=load_balancers,
            network_configuration=network_configuration,
            ordered_placement_strategies=ordered_placement_strategies,
            placement_constraints=placement_constraints,
            platform_version=platform_version,
            propagate_tags=propagate_tags,
            scheduling_strategy=scheduling_strategy,
            service_connect_configuration=service_connect_configuration,
            service_registries=service_registries,
            tags=tags,
            task_definition=task_definition,
            triggers=triggers,
            wait_for_steady_state=wait_for_steady_state,
        )

    def place(self, cluster):
        self.kwargs["cluster"] = cluster

    def with_domain(
        self,
        vpc_id,
        domain,
        load_balancer: LoadBalancerFactory,
        container,
        port,
        zone_id,
        certificate=None,
        dns_opts=None,
    ):
        target_group_factory = TargetGroupFactory(
            self.resource_name, port=port, protocol="HTTP", vpc_id=vpc_id, deregistration_delay=120, target_type="ip"
        )

        target_group = target_group_factory.create()

        rule_factory = load_balancer.create_https_rule(
            self.resource_name, target_group=target_group, domain_name=domain
        )
        rule_factory.create()

        self.kwargs["load_balancers"] = [
            ecs.ServiceLoadBalancerArgs(
                target_group_arn=target_group.arn,
                container_name=container or self.resource_name,
                container_port=port or 80,
            )
        ]

        zone_factory = ZoneFactory(resource_name="zone", resource_id=zone_id)
        alb = load_balancer.get()
        zone_factory.alias(
            name=domain,
            target_zone=alb.zone_id,
            target_record=alb.dns_name,
            resource_name=self.identifier.add("alias"),
            opts=dns_opts,
        )
        # zone_factory.create()

        if certificate:
            lb.ListenerCertificate(
                "service_certificate",
                listener_arn=load_balancer.get_https_listener().arn,
                certificate_arn=certificate.arn,
            )

    def create(self) -> ecs.Service:
        self.setdefault(
            "capacity_provider_strategies",
            [ecs.ServiceCapacityProviderStrategyArgs(base=1, capacity_provider="FARGATE", weight=100)],
        )

        if not self.kwargs.get("desired_count", None):
            self.kwargs["desired_count"] = 1
        return self._create(resource_type=ecs.Service)


class TaskDefinitionFactory(BaseFactory):
    @overload
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        container_definitions: Optional[pulumi.Input[str]] = None,
        cpu: Optional[pulumi.Input[str]] = None,
        ephemeral_storage: Optional[pulumi.Input[pulumi.InputType["TaskDefinitionEphemeralStorageArgs"]]] = None,
        execution_role_arn: Optional[pulumi.Input[str]] = None,
        family: Optional[pulumi.Input[str]] = None,
        inference_accelerators: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["TaskDefinitionInferenceAcceleratorArgs"]]]]
        ] = None,
        ipc_mode: Optional[pulumi.Input[str]] = None,
        memory: Optional[pulumi.Input[str]] = None,
        network_mode: Optional[pulumi.Input[str]] = None,
        pid_mode: Optional[pulumi.Input[str]] = None,
        placement_constraints: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["TaskDefinitionPlacementConstraintArgs"]]]]
        ] = None,
        proxy_configuration: Optional[pulumi.Input[pulumi.InputType["TaskDefinitionProxyConfigurationArgs"]]] = None,
        requires_compatibilities: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        runtime_platform: Optional[pulumi.Input[pulumi.InputType["TaskDefinitionRuntimePlatformArgs"]]] = None,
        skip_destroy: Optional[pulumi.Input[bool]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        task_role_arn: Optional[pulumi.Input[str]] = None,
        volumes: Optional[pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["TaskDefinitionVolumeArgs"]]]]] = None,
        __props__=None,
    ):
        pass

    @overload
    def __init__(self, name, resource_id):
        pass

    def __init__(self, name: str, *args, **kwargs):
        super().__init__(name, **kwargs)
        if args:
            self.id = args[0]

        self.service = None
        self.role = None
        self.secret = None
        self.containers = []

    def with_container(self, container: ContainerDefinition):
        self.containers.append(container)
        return self.containers

    def with_secret(self, secret):
        self.secret = secret

    def with_service(
        self,
        resource_name: str = None,
        opts: Optional[pulumi.ResourceOptions] = None,
        alarms: Optional[pulumi.Input[pulumi.InputType["ServiceAlarmsArgs"]]] = None,
        capacity_provider_strategies: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ServiceCapacityProviderStrategyArgs"]]]]
        ] = None,
        cluster: Optional[pulumi.Input[str]] = None,
        deployment_circuit_breaker: Optional[
            pulumi.Input[pulumi.InputType["ServiceDeploymentCircuitBreakerArgs"]]
        ] = None,
        deployment_controller: Optional[pulumi.Input[pulumi.InputType["ServiceDeploymentControllerArgs"]]] = None,
        deployment_maximum_percent: Optional[pulumi.Input[int]] = None,
        deployment_minimum_healthy_percent: Optional[pulumi.Input[int]] = None,
        desired_count: Optional[pulumi.Input[int]] = None,
        enable_ecs_managed_tags: Optional[pulumi.Input[bool]] = None,
        enable_execute_command: Optional[pulumi.Input[bool]] = None,
        force_new_deployment: Optional[pulumi.Input[bool]] = None,
        health_check_grace_period_seconds: Optional[pulumi.Input[int]] = None,
        iam_role: Optional[pulumi.Input[str]] = None,
        launch_type: Optional[pulumi.Input[str]] = None,
        load_balancers: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ServiceLoadBalancerArgs"]]]]
        ] = None,
        network_configuration: Optional[pulumi.Input[pulumi.InputType["ServiceNetworkConfigurationArgs"]]] = None,
        ordered_placement_strategies: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ServiceOrderedPlacementStrategyArgs"]]]]
        ] = None,
        placement_constraints: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ServicePlacementConstraintArgs"]]]]
        ] = None,
        platform_version: Optional[pulumi.Input[str]] = None,
        propagate_tags: Optional[pulumi.Input[str]] = None,
        scheduling_strategy: Optional[pulumi.Input[str]] = None,
        service_connect_configuration: Optional[
            pulumi.Input[pulumi.InputType["ServiceServiceConnectConfigurationArgs"]]
        ] = None,
        service_registries: Optional[pulumi.Input[pulumi.InputType["ServiceServiceRegistriesArgs"]]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        task_definition: Optional[pulumi.Input[str]] = None,
        triggers: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        wait_for_steady_state: Optional[pulumi.Input[bool]] = None,
        __props__=None,
    ) -> ServiceFactory:
        self.service = ServiceFactory(
            resource_name=resource_name or self.identifier.add("service"),
            opts=opts,
            alarms=alarms,
            capacity_provider_strategies=capacity_provider_strategies,
            cluster=cluster,
            deployment_circuit_breaker=deployment_circuit_breaker,
            deployment_controller=deployment_controller,
            deployment_maximum_percent=deployment_maximum_percent,
            deployment_minimum_healthy_percent=deployment_minimum_healthy_percent,
            desired_count=desired_count,
            enable_ecs_managed_tags=enable_ecs_managed_tags,
            enable_execute_command=enable_execute_command,
            force_new_deployment=force_new_deployment,
            health_check_grace_period_seconds=health_check_grace_period_seconds,
            iam_role=iam_role,
            launch_type=launch_type,
            load_balancers=load_balancers,
            network_configuration=network_configuration,
            ordered_placement_strategies=ordered_placement_strategies,
            placement_constraints=placement_constraints,
            platform_version=platform_version,
            propagate_tags=propagate_tags,
            scheduling_strategy=scheduling_strategy,
            service_connect_configuration=service_connect_configuration,
            service_registries=service_registries,
            tags=tags,
            task_definition=task_definition,
            triggers=triggers,
            wait_for_steady_state=wait_for_steady_state,
            __props__=__props__,
        )
        return self.service

    def with_role(
        self,
        opts: Optional[pulumi.ResourceOptions] = None,
        assume_role_policy: Optional[pulumi.Input[str]] = None,
        description: Optional[pulumi.Input[str]] = None,
        force_detach_policies: Optional[pulumi.Input[bool]] = None,
        inline_policies: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["RoleInlinePolicyArgs"]]]]
        ] = None,
        managed_policy_arns: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        max_session_duration: Optional[pulumi.Input[int]] = None,
        name: Optional[pulumi.Input[str]] = None,
        name_prefix: Optional[pulumi.Input[str]] = None,
        path: Optional[pulumi.Input[str]] = None,
        permissions_boundary: Optional[pulumi.Input[str]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        __props__=None,
    ):
        if not assume_role_policy:
            assume_role_policy = {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Sid": "",
                        "Effect": "Allow",
                        "Principal": {"Service": "ecs-tasks.amazonaws.com"},
                        "Action": "sts:AssumeRole",
                    }
                ],
            }

        if not isinstance(managed_policy_arns, list):
            managed_policy_arns = []

        if "arn:aws:iam::aws:policy/service-role/AmazonECSTaskExecutionRolePolicy" not in managed_policy_arns:
            managed_policy_arns.append("arn:aws:iam::aws:policy/service-role/AmazonECSTaskExecutionRolePolicy")

        if not isinstance(inline_policies, list):
            inline_policies = []

        if self.secret:
            inline_policies.append(
                iam.RoleInlinePolicyArgs(
                    name="allow_secrets",
                    policy=json.dumps(
                        {
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Version": "2012-10-17",
                                    "Statement": [
                                        {
                                            "Effect": "Allow",
                                            "Action": [
                                                "secretsmanager:GetSecretValue",
                                            ],
                                            "Resource": [
                                                self.secret.arn,
                                            ],
                                        }
                                    ],
                                }
                            ],
                        }
                    ),
                ),
            )

        inline_policies_names = [policy.get("name") for policy in inline_policies]
        if "cloudwatch" not in inline_policies_names:
            log_group = cloudwatch.LogGroup(
                self.identifier.snake, name="/ecs/" + self.identifier.spinal, retention_in_days=365
            )
            inline_policies.append(
                RoleInlinePolicyArgs(
                    name="cloudwatch",
                    policy=log_group.arn.apply(
                        lambda arn: json.dumps(
                            {
                                "Version": "2012-10-17",
                                "Statement": [
                                    {
                                        "Action": [
                                            "logs:CreateLogGroup",
                                            "logs:CreateLogStream",
                                            "logs:PutLogEvents",
                                            "logs:DescribeLogStreams",
                                        ],
                                        "Effect": "Allow",
                                        "Resource": [arn],
                                    },
                                ],
                            }
                        )
                    ),
                )
            )

        self.role = Role(
            name=name or self.identifier.add("role"),
            opts=opts,
            assume_role_policy=assume_role_policy,
            description=description,
            force_detach_policies=force_detach_policies,
            inline_policies=inline_policies,
            managed_policy_arns=managed_policy_arns,
            max_session_duration=max_session_duration,
            name_prefix=name_prefix,
            path=path,
            permissions_boundary=permissions_boundary,
            tags=tags,
            __props__=__props__,
        )
        self.kwargs["execution_role_arn"] = self.role.arn
        return self.role

    def create(self) -> ecs.TaskDefinition:
        self.kwargs["container_definitions"] = pulumi.Output.json_dumps(self.containers)
        self.kwargs.setdefault("family", str(self.resource_name))
        self.kwargs.setdefault("requires_compatibilities", ["FARGATE"])
        self.kwargs.setdefault("network_mode", "awsvpc")

        task = self._create(resource_type=ecs.TaskDefinition)
        if self.service:
            self.service.kwargs["task_definition"] = task.id
            self.service.create()

        return task
