from typing import Optional, Sequence, Mapping

from pulumi_aws import elasticache

from cloudlib.modules.ec2 import SecurityGroup
from cloudlib.modules.resource import BaseFactory
import pulumi

from cloudlib.utils.names import Identifier, IdentifierV2


class ClusterFactory(BaseFactory):
    def __init__(
        self,
        resource_name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        apply_immediately: Optional[pulumi.Input[bool]] = None,
        auto_minor_version_upgrade: Optional[pulumi.Input[str]] = None,
        availability_zone: Optional[pulumi.Input[str]] = None,
        az_mode: Optional[pulumi.Input[str]] = None,
        cluster_id: Optional[pulumi.Input[str]] = None,
        engine: Optional[pulumi.Input[str]] = None,
        engine_version: Optional[pulumi.Input[str]] = None,
        final_snapshot_identifier: Optional[pulumi.Input[str]] = None,
        ip_discovery: Optional[pulumi.Input[str]] = None,
        log_delivery_configurations: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ClusterLogDeliveryConfigurationArgs"]]]]
        ] = None,
        maintenance_window: Optional[pulumi.Input[str]] = None,
        network_type: Optional[pulumi.Input[str]] = None,
        node_type: Optional[pulumi.Input[str]] = None,
        notification_topic_arn: Optional[pulumi.Input[str]] = None,
        num_cache_nodes: Optional[pulumi.Input[int]] = None,
        outpost_mode: Optional[pulumi.Input[str]] = None,
        parameter_group_name: Optional[pulumi.Input[str]] = None,
        port: Optional[pulumi.Input[int]] = None,
        preferred_availability_zones: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        preferred_outpost_arn: Optional[pulumi.Input[str]] = None,
        replication_group_id: Optional[pulumi.Input[str]] = None,
        security_group_ids: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        snapshot_arns: Optional[pulumi.Input[str]] = None,
        snapshot_name: Optional[pulumi.Input[str]] = None,
        snapshot_retention_limit: Optional[pulumi.Input[int]] = None,
        snapshot_window: Optional[pulumi.Input[str]] = None,
        subnet_group_name: Optional[pulumi.Input[str]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
    ):
        super().__init__(
            resource_name=resource_name,
            opts=opts,
            apply_immediately=apply_immediately,
            auto_minor_version_upgrade=auto_minor_version_upgrade,
            availability_zone=availability_zone,
            az_mode=az_mode,
            cluster_id=cluster_id,
            engine=engine,
            engine_version=engine_version,
            final_snapshot_identifier=final_snapshot_identifier,
            ip_discovery=ip_discovery,
            log_delivery_configurations=log_delivery_configurations,
            maintenance_window=maintenance_window,
            network_type=network_type,
            node_type=node_type,
            notification_topic_arn=notification_topic_arn,
            num_cache_nodes=num_cache_nodes,
            outpost_mode=outpost_mode,
            parameter_group_name=parameter_group_name,
            port=port,
            preferred_availability_zones=preferred_availability_zones,
            preferred_outpost_arn=preferred_outpost_arn,
            replication_group_id=replication_group_id,
            security_group_ids=security_group_ids,
            snapshot_arns=snapshot_arns,
            snapshot_name=snapshot_name,
            snapshot_retention_limit=snapshot_retention_limit,
            snapshot_window=snapshot_window,
            subnet_group_name=subnet_group_name,
            tags=tags,
        )
        self.allow_cidr = "0.0.0.0/0"

    def place(self, subnet_ids, vpc_id, allow_cidr=None):
        self.subnet_group = elasticache.SubnetGroup(
            self.resource_name, name=IdentifierV2(self.resource_name).spinal, subnet_ids=subnet_ids
        )
        self.vpc_id = vpc_id
        self.set_default("subnet_group_name", self.subnet_group.name)
        if allow_cidr:
            self.allow_cidr = allow_cidr

    def create(self) -> elasticache.Cluster:
        self.set_default("cluster_id", self.resource_name)
        self.set_default("engine", "redis")
        self.set_default("engine_version", "7.0")
        self.set_default("node_type", "cache.t2.medium")
        self.set_default("num_cache_nodes", 1)
        self.set_default("snapshot_retention_limit", 14)
        self.set_default("port", 6379)

        if not self.kwargs.get("security_group_ids"):
            self.set_default(
                "security_group_ids",
                [
                    SecurityGroup(
                        name=f"elasticache_{self.resource_name}",
                        vpc_id=self.vpc_id,
                        egress=[
                            dict(
                                from_port=0,
                                to_port=0,
                                protocol="-1",
                                cidr_blocks=["0.0.0.0/0"],
                            )
                        ],
                        ingress=[
                            dict(
                                from_port=6379,
                                to_port=6379,
                                protocol="TCP",
                                cidr_blocks=[str(self.allow_cidr)],
                            )
                        ],
                    ).id
                ],
            )

        return self._create(resource_type=elasticache.Cluster)
