import json
from typing import Optional, Sequence, Mapping

import pulumi
from pulumi_aws import iam

from cloudlib.modules.resource import BaseResourceMixin, BaseFactory, AttributeResolver
from cloudlib.utils.names import IdentifierV2


class Role(iam.Role, BaseResourceMixin):
    def __init__(
        self,
        name,
        opts: Optional[pulumi.ResourceOptions] = None,
        assume_role_policy: Optional[pulumi.Input[str]] = None,
        description: Optional[pulumi.Input[str]] = None,
        force_detach_policies: Optional[pulumi.Input[bool]] = None,
        inline_policies: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["RoleInlinePolicyArgs"]]]]
        ] = None,
        managed_policy_arns: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        max_session_duration: Optional[pulumi.Input[int]] = None,
        name_prefix: Optional[pulumi.Input[str]] = None,
        path: Optional[pulumi.Input[str]] = None,
        permissions_boundary: Optional[pulumi.Input[str]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        __props__=None,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=iam.Role,
            resource_name=name,
            name=IdentifierV2(name).pascal,
            with_prefix=True,
            opts=opts,
            assume_role_policy=assume_role_policy,
            description=description,
            force_detach_policies=force_detach_policies,
            inline_policies=inline_policies,
            managed_policy_arns=managed_policy_arns,
            max_session_duration=max_session_duration,
            name_prefix=name_prefix,
            path=path,
            permissions_boundary=permissions_boundary,
            tags=tags,
        )


class Statement:
    def __init__(
        self,
        actions: Optional[Sequence[str]] = None,
        conditions: Optional[Sequence["GetPolicyDocumentStatementConditionArgs"]] = None,
        effect: Optional[str] = None,
        not_actions: Optional[Sequence[str]] = None,
        not_principals: Optional[Sequence["GetPolicyDocumentStatementNotPrincipalArgs"]] = None,
        not_resources: Optional[Sequence[str]] = None,
        principals: Optional[Sequence["GetPolicyDocumentStatementPrincipalArgs"]] = None,
        resources: Optional[Sequence[str]] = None,
        sid: Optional[str] = None,
    ):
        self.actions = actions
        self.conditions = conditions
        self.effect = effect
        self.not_actions = not_actions
        self.not_resources = not_resources
        self.not_principals = not_principals
        self.principals = principals
        self.resources = resources if isinstance(resources, list) else [resources]
        self.sid = sid

    def create(self):
        return {
            k: v
            for k, v in dict(
                Action=self.actions,
                Condition=self.conditions,
                Effect=self.effect,
                NotAction=self.not_actions,
                NotPrincipal=self.not_principals,
                NotResource=self.not_resources,
                Principal=self.principals,
                Resource=[
                    resource.resolve() if isinstance(resource, AttributeResolver) else resource
                    for resource in self.resources
                ],
                Sid=self.sid,
            ).items()
            if v
        }


class Policy:
    def __init__(self, name, statements):
        self.name = name
        self.statements = statements if isinstance(statements, list) else [statements]

    def create(self):
        return dict(
            name=self.name,
            policy=json.dumps(
                {"Version": "2012-10-17", "Statement": [statement.create() for statement in self.statements]}
            ),
        )


class RoleFactory(BaseFactory):
    def __init__(
        self,
        name,
        opts: Optional[pulumi.ResourceOptions] = None,
        assume_role_policy: Optional[pulumi.Input[str]] = None,
        description: Optional[pulumi.Input[str]] = None,
        force_detach_policies: Optional[pulumi.Input[bool]] = None,
        inline_policies: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["RoleInlinePolicyArgs"]]]]
        ] = None,
        managed_policy_arns: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        max_session_duration: Optional[pulumi.Input[int]] = None,
        name_prefix: Optional[pulumi.Input[str]] = None,
        path: Optional[pulumi.Input[str]] = None,
        permissions_boundary: Optional[pulumi.Input[str]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
    ):
        super().__init__(
            resource_name=name,
            name=IdentifierV2(name).pascal,
            opts=opts,
            assume_role_policy=assume_role_policy,
            description=description,
            force_detach_policies=force_detach_policies,
            inline_policies=inline_policies,
            managed_policy_arns=managed_policy_arns,
            max_session_duration=max_session_duration,
            name_prefix=name_prefix,
            path=path,
            permissions_boundary=permissions_boundary,
            tags=tags,
        )
        self.inline_policies_items = []

    def with_inline_policy(self, policy: Policy):
        self.inline_policies_items.append(policy)

    def create(self):
        self.setdefault("inline_policies", [policy.create() for policy in self.inline_policies_items])

        return self._create(iam.Role)
