import json
from typing import Optional, Mapping, overload

import pulumi
from pulumi_aws import kms, get_caller_identity, iam
from pulumi_aws.kms import KeyArgs

from cloudlib.modules.iam import Statement, Policy
from cloudlib.modules.resource import BaseResourceMixin, BaseFactory
from cloudlib.utils.names import IdentifierV2


class Key(kms.Key, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        bypass_policy_lockout_safety_check: Optional[pulumi.Input[bool]] = None,
        custom_key_store_id: Optional[pulumi.Input[str]] = None,
        customer_master_key_spec: Optional[pulumi.Input[str]] = None,
        deletion_window_in_days: Optional[pulumi.Input[int]] = None,
        description: Optional[pulumi.Input[str]] = None,
        enable_key_rotation: Optional[pulumi.Input[bool]] = None,
        is_enabled: Optional[pulumi.Input[bool]] = None,
        key_usage: Optional[pulumi.Input[str]] = None,
        multi_region: Optional[pulumi.Input[bool]] = None,
        policy: Optional[pulumi.Input[str]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        __props__=None,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=kms.Key,
            resource_name=name,
            opts=opts,
            bypass_policy_lockout_safety_check=bypass_policy_lockout_safety_check,
            custom_key_store_id=custom_key_store_id,
            customer_master_key_spec=customer_master_key_spec,
            deletion_window_in_days=deletion_window_in_days,
            description=description,
            enable_key_rotation=enable_key_rotation,
            is_enabled=is_enabled,
            key_usage=key_usage,
            multi_region=multi_region,
            policy=policy,
            tags=tags,
            __props__=__props__,
        )

    def create_alias(self, name):
        pass


class AliasFactory(BaseFactory):
    def __init__(
        self,
        opts: Optional[pulumi.ResourceOptions] = None,
        name: Optional[pulumi.Input[str]] = None,
        name_prefix: Optional[pulumi.Input[str]] = None,
        target_key_id: Optional[pulumi.Input[str]] = None,
    ):
        if not name.startswith("alias/"):
            name = f"alias/{IdentifierV2(name).spinal}"

        super().__init__(resource_name=name, name=name, name_prefix=name_prefix, target_key_id=target_key_id, opts=opts)

    def create(self):
        return self._create(kms.Alias)


class KeyFactory(BaseFactory):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        bypass_policy_lockout_safety_check: Optional[pulumi.Input[bool]] = None,
        custom_key_store_id: Optional[pulumi.Input[str]] = None,
        customer_master_key_spec: Optional[pulumi.Input[str]] = None,
        deletion_window_in_days: Optional[pulumi.Input[int]] = None,
        description: Optional[pulumi.Input[str]] = None,
        enable_key_rotation: Optional[pulumi.Input[bool]] = None,
        is_enabled: Optional[pulumi.Input[bool]] = None,
        key_usage: Optional[pulumi.Input[str]] = None,
        multi_region: Optional[pulumi.Input[bool]] = None,
        policy: Optional[pulumi.Input[str]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
    ):
        super().__init__(
            resource_name=name,
            opts=opts,
            bypass_policy_lockout_safety_check=bypass_policy_lockout_safety_check,
            custom_key_store_id=custom_key_store_id,
            customer_master_key_spec=customer_master_key_spec,
            deletion_window_in_days=deletion_window_in_days,
            description=description,
            enable_key_rotation=enable_key_rotation,
            is_enabled=is_enabled,
            key_usage=key_usage,
            multi_region=multi_region,
            policy=policy,
            tags=tags,
        )

        self.account_id = get_caller_identity().account_id

        self.alias_factory = None
        self.policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "RootPermissions",
                    "Effect": "Allow",
                    "Principal": {"AWS": f"arn:aws:iam::{self.account_id}:root"},
                    "Action": "kms:*",
                    "Resource": "*",
                }
            ],
        }

    def with_alias(self, alias):
        self.alias_factory = AliasFactory(name=alias, target_key_id=self.ref("key_id"))
        return self.alias_factory

    def usage_policy(self):
        return Policy(
            "kms_key_usage",
            Statement(effect="Allow", actions=["kms:Encrypt", "kms:Decrypt"], resources=[self.ref("arn")]),
        )

    def create(self):
        self.setdefault("enable_key_rotation", True)
        self.setdefault("policy", json.dumps(self.policy))

        key: kms.Key = self._create(kms.Key)

        if self.alias_factory:
            self.alias_factory.create()

        return key

    def allow_usage(self, resource: BaseFactory, suffix=""):
        statements = resource.kms_policy(suffix=suffix)
        if not isinstance(statements, list):
            statements = [statements]

        for statement in statements:
            self.policy["Statement"].append(statement)
