import inflection
from cloudlib.utils.names import IdentifierV2
from pulumi_aws import ec2, get_region, get_caller_identity


region_id = get_region().id
account_id = get_caller_identity().account_id


class BaseResourceMixin:
    def apply_tags(self, args, args_type, **tags):
        if not isinstance(args, args_type):
            args = args_type()

        if not isinstance(args.tags, dict):
            args.tags = {}

        for name, value in tags.items():
            args.tags.setdefault(name, value)

        return args

    def __init__(self, parent, resource_name, with_prefix=False, **kwargs):
        if "identifier" not in kwargs:
            _id = self.identifier = IdentifierV2(resource_name)
        else:
            _id = self._identifier = IdentifierV2(resource_name)

        if "tags" in kwargs:
            kwargs["tags"] = kwargs["tags"] or {}
            kwargs["tags"].setdefault("Name", _id.spinal)

        parent.__init__(self, _id.snake, **kwargs)


class AttributeResolver:
    def __init__(self, obj, attribute):
        self.obj = obj
        self.attribute = attribute

    def resolve(self):
        return getattr(self.obj.resource, self.attribute)


class BaseFactory:
    def __init__(self, resource_name, **kwargs):
        self.identifier = IdentifierV2(resource_name)
        self.resource_name = self.identifier.snake
        self.kwargs = kwargs

    def _create(self, resource_type):
        kwargs = self.kwargs
        resource_name = self.resource_name
        if "identifier" not in kwargs:
            _id = self.identifier = IdentifierV2(resource_name)
        else:
            _id = IdentifierV2(resource_name)

        kwargs.setdefault("resource_name", _id.snake)

        if "tags" in kwargs:
            kwargs["tags"] = kwargs["tags"] or {}
            kwargs["tags"].setdefault("Name", _id.spinal)

        for arg, value in kwargs.items():
            if isinstance(value, AttributeResolver):
                kwargs[arg] = value.resolve()

        self.resource = resource_type(**kwargs)
        return self.resource

    @property
    def name(self):
        return self.kwargs.get("name")

    def set_default(self, name, value):
        if not self.kwargs.get(name):
            self.kwargs[name] = value

    def setdefault(self, name, value):
        self.set_default(name, value)

    def ref(self, name):
        return AttributeResolver(self, name)

    def service(self):
        return None

    def kms_policy(self, suffix=""):
        return {
            "Effect": "Allow",
            "Principal": {"Service": f"{self.service()}"},
            "Action": [
                "kms:Encrypt*",
                "kms:Decrypt*",
                "kms:ReEncrypt*",
                "kms:GenerateDataKey*",
                "kms:Describe*",
            ],
            "Resource": "*",
            "Condition": {"ArnLike": {"kms:EncryptionContext:aws:logs:arn": self.arn() + suffix}},
        }

    def arn(self, resource_type, resource_id, region=None, account=None):
        return f"arn:aws:{resource_type}:{region or region_id}:{account or account_id}:{resource_id}"

    def apply_tags(self, args, args_type, **tags):
        if not isinstance(args, args_type):
            args = args_type()

        if not isinstance(args.tags, dict):
            args.tags = {}

        for name, value in tags.items():
            args.tags.setdefault(name, value)

        return args
