from typing import Optional, Mapping, Sequence, Union

import pulumi
from pulumi_aws import route53, acm

from cloudlib.modules.resource import BaseResourceMixin, BaseFactory, AttributeResolver
from cloudlib.utils.names import IdentifierV2


class Zone(route53.Zone, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        comment: Optional[pulumi.Input[str]] = None,
        delegation_set_id: Optional[pulumi.Input[str]] = None,
        force_destroy: Optional[pulumi.Input[bool]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        vpcs: Optional[pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ZoneVpcArgs"]]]]] = None,
        __props__=None,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=route53.Zone,
            resource_name=name,
            opts=opts,
            comment=comment,
            delegation_set_id=delegation_set_id,
            force_destroy=force_destroy,
            name=name,
            tags=tags,
            vpcs=vpcs,
            __props__=__props__,
        )

    def share(self, vpc_id):
        return VpcAssociationAuthorization(name=self.identifier.add(f"share"), vpc_id=vpc_id, zone_id=self.zone_id)

    def a_record(self, name, target, **kwargs):
        kwargs["zone_id"] = self.zone_id
        kwargs["type"] = route53.RecordType.A
        kwargs["records"] = target if isinstance(target, list) else [target]
        kwargs["ttl"] = 60
        return Record(name=name, **kwargs)


class ZoneFactory(BaseFactory):
    def __init__(
        self,
        resource_name: str,
        resource_id=None,
        opts: Optional[pulumi.ResourceOptions] = None,
        comment: Optional[pulumi.Input[str]] = None,
        delegation_set_id: Optional[pulumi.Input[str]] = None,
        force_destroy: Optional[pulumi.Input[bool]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        vpcs: Optional[pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ZoneVpcArgs"]]]]] = None,
        __props__=None,
    ):
        self.zone_id = None
        if resource_id:
            self.zone_id = resource_id
        else:
            super().__init__(
                resource_name=resource_name,
                opts=opts,
                comment=comment,
                delegation_set_id=delegation_set_id,
                force_destroy=force_destroy,
                name=resource_name,
                tags=tags,
                vpcs=vpcs,
                __props__=__props__,
            )

        self.records = []

    def record(self, **kwargs) -> route53.Record:
        if "resource_name" not in kwargs:
            kwargs["resource_name"] = IdentifierV2(kwargs.get("name") or "").snake

        if self.zone_id:
            return route53.Record(zone_id=self.zone_id, **kwargs)
        else:
            return self.records.append({**{"zone_id": self.ref("zone_id")}, **kwargs})

    def alias(
        self,
        name,
        target_zone=None,
        target_record=None,
        resource_name=None,
        records=None,
        set_identifier=None,
        weighted_routing_policies=None,
        opts=None,
    ):
        if not records:
            records = [dict(name=target_record, zone_id=target_zone, evaluate_target_health=False)]

        self.record(
            resource_name=resource_name or f"{name}_alias",
            weighted_routing_policies=weighted_routing_policies,
            set_identifier=set_identifier,
            name=name,
            type="A",
            aliases=[route53.RecordAliasArgs(**record) for record in records],
            opts=opts,
        )

    def validate_certificate(self, certificate: acm.Certificate, opts=None):
        def validation_record(options):
            records = []
            for idx, option in enumerate(options):
                records.append(
                    self.record(
                        resource_name=f"certificate_validation_{idx}",
                        name=option.resource_record_name,
                        records=[option.resource_record_value],
                        ttl=60,
                        type=option.resource_record_type,
                        opts=opts,
                    ).fqdn
                )

            acm.CertificateValidation(
                resource_name="certificate_validation",
                certificate_arn=certificate.arn,
                validation_record_fqdns=records,
            )

        certificate.domain_validation_options.apply(validation_record)

    def create(self) -> Zone:
        if self.kwargs.get("vpcs"):
            self.resource_name = self.resource_name + "-private"

        zone = self._create(resource_type=route53.Zone)
        for record in self.records:
            record["zone_id"] = zone.id
            route53.Record(**record)

        return zone


class Record(route53.Record, BaseResourceMixin):
    def __init__(
        self,
        parent=route53.Record,
        opts: Optional[pulumi.ResourceOptions] = None,
        aliases: Optional[pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["RecordAliasArgs"]]]]] = None,
        allow_overwrite: Optional[pulumi.Input[bool]] = None,
        cidr_routing_policy: Optional[pulumi.Input[pulumi.InputType["RecordCidrRoutingPolicyArgs"]]] = None,
        failover_routing_policies: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["RecordFailoverRoutingPolicyArgs"]]]]
        ] = None,
        geolocation_routing_policies: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["RecordGeolocationRoutingPolicyArgs"]]]]
        ] = None,
        health_check_id: Optional[pulumi.Input[str]] = None,
        latency_routing_policies: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["RecordLatencyRoutingPolicyArgs"]]]]
        ] = None,
        multivalue_answer_routing_policy: Optional[pulumi.Input[bool]] = None,
        name: Optional[pulumi.Input[str]] = None,
        records: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        set_identifier: Optional[pulumi.Input[str]] = None,
        ttl: Optional[pulumi.Input[int]] = None,
        type: Optional[pulumi.Input[Union[str, "RecordType"]]] = None,
        weighted_routing_policies: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["RecordWeightedRoutingPolicyArgs"]]]]
        ] = None,
        zone_id: Optional[pulumi.Input[str]] = None,
        __props__=None,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=route53.Record,
            resource_name=name,
            opts=opts,
            aliases=aliases,
            allow_overwrite=allow_overwrite,
            cidr_routing_policy=cidr_routing_policy,
            failover_routing_policies=failover_routing_policies,
            geolocation_routing_policies=geolocation_routing_policies,
            health_check_id=health_check_id,
            latency_routing_policies=latency_routing_policies,
            multivalue_answer_routing_policy=multivalue_answer_routing_policy,
            name=name,
            records=records,
            set_identifier=set_identifier,
            ttl=ttl,
            type=type,
            weighted_routing_policies=weighted_routing_policies,
            zone_id=zone_id,
            __props__=__props__,
        )


class ResolverEndpoint(route53.ResolverEndpoint, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        direction: Optional[pulumi.Input[str]] = None,
        ip_addresses: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ResolverEndpointIpAddressArgs"]]]]
        ] = None,
        security_group_ids: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        __props__=None,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=route53.ResolverEndpoint,
            resource_name=name,
            opts=opts,
            direction=direction,
            ip_addresses=ip_addresses,
            security_group_ids=security_group_ids,
            name=IdentifierV2(name).spinal,
            tags=tags,
        )


class VpcAssociationAuthorization(route53.VpcAssociationAuthorization, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        vpc_id: Optional[pulumi.Input[str]] = None,
        vpc_region: Optional[pulumi.Input[str]] = None,
        zone_id: Optional[pulumi.Input[str]] = None,
        __props__=None,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=route53.VpcAssociationAuthorization,
            resource_name=name,
            opts=opts,
            vpc_id=vpc_id,
            vpc_region=vpc_region,
            zone_id=zone_id,
            __props__=__props__,
        )


class ZoneAssociation(route53.ZoneAssociation, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        vpc_id: Optional[pulumi.Input[str]] = None,
        vpc_region: Optional[pulumi.Input[str]] = None,
        zone_id: Optional[pulumi.Input[str]] = None,
        __props__=None,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=route53.ZoneAssociation,
            resource_name=name,
            opts=opts,
            vpc_id=vpc_id,
            vpc_region=vpc_region,
            zone_id=zone_id,
            __props__=__props__,
        )


class ResolverRule(route53.ResolverRule, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        domain_name: Optional[pulumi.Input[str]] = None,
        resolver_endpoint_id: Optional[pulumi.Input[str]] = None,
        rule_type: Optional[pulumi.Input[str]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        target_ips: Optional[pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["ResolverRuleTargetIpArgs"]]]]] = None,
        __props__=None,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=route53.ResolverRule,
            resource_name=name,
            name=IdentifierV2(name).spinal,
            opts=opts,
            domain_name=domain_name,
            resolver_endpoint_id=resolver_endpoint_id,
            rule_type=rule_type,
            tags=tags,
            target_ips=target_ips,
            __props__=__props__,
        )


class ResolverRuleAssociation(route53.ResolverRuleAssociation, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        resolver_rule_id: Optional[pulumi.Input[str]] = None,
        vpc_id: Optional[pulumi.Input[str]] = None,
        __props__=None,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=route53.ResolverRuleAssociation,
            resource_name=name,
            opts=opts,
            resolver_rule_id=resolver_rule_id,
            vpc_id=vpc_id,
            __props__=__props__,
        )
