from typing import Optional, Sequence, overload, Mapping, Union

import pulumi
from pulumi_aws import s3 as s3classic, get_caller_identity, get_region
from pulumi_aws import iam
from pulumi_aws.s3 import BucketV2Args
from pulumi_aws_native import s3
from pulumi_aws_native.s3 import BucketArgs

from cloudlib.modules.resource import BaseFactory
from cloudlib.utils.names import regional_domain, IdentifierV2
from pulumi import ResourceOptions


class Bucket(s3.Bucket):
    def __init__(__self__, resource_name: str, namespace="epsy", *args, **kwargs):
        kwargs.setdefault("access_control", s3.BucketAccessControl.PRIVATE)
        kwargs.setdefault(
            "public_access_block_configuration",
            s3.BucketPublicAccessBlockConfigurationArgs(
                block_public_acls=True, block_public_policy=True, ignore_public_acls=True, restrict_public_buckets=True
            ),
        )
        kwargs.setdefault(
            "versioning_configuration",
            s3.BucketVersioningConfigurationArgs(status=s3.BucketVersioningConfigurationStatus.ENABLED),
        )

        kwargs["bucket_name"] = regional_domain(namespace, sub=kwargs.get("bucket_name", None))
        kwargs.setdefault("opts", ResourceOptions(replace_on_changes=["bucket_name"], retain_on_delete=True))

        super().__init__(resource_name, *args, **kwargs)


class EncryptedBucket(Bucket):
    def __init__(__self__, resource_name: str, namespace="epsy", *args, **kwargs):
        kwargs.setdefault("bucket_encryption", s3.BucketEncryptionArgs())
        super().__init__(resource_name, namespace=namespace, *args, **kwargs)


class BucketPolicyFactory:
    def __init__(self, extra_statements=None):
        self.extra_statements = extra_statements or []

    def create(self, bucket, extra_statements=None):
        return iam.get_policy_document_output(
            statements=[
                iam.GetPolicyDocumentStatementArgs(
                    sid="AllowSSLRequestsOnly",
                    effect="Deny",
                    principals=[iam.GetPolicyDocumentStatementPrincipalArgs(identifiers=["*"], type="*")],
                    actions=["s3:*"],
                    resources=[f"arn:aws:s3:::{bucket}", f"arn:aws:s3:::{bucket}/*"],
                    conditions=[
                        iam.GetPolicyDocumentStatementConditionArgs(
                            test="Bool",
                            variable="aws:SecureTransport",
                            values=[False],
                        )
                    ],
                )
            ]
            + (extra_statements or [])
            + (self.extra_statements or [])
        )


class S3BucketAccessLogPolicyFactory(BucketPolicyFactory):
    def __init__(self, accounts=None):
        self.accounts = accounts
        super().__init__()

    def create(self, bucket):
        identity = get_caller_identity()
        accounts = self.accounts or [identity.account_id]

        statements = []

        for account_id in accounts:
            statements.append(
                iam.GetPolicyDocumentStatementArgs(
                    sid=f"ACLCheck_{account_id}",
                    effect="Allow",
                    principals=[
                        iam.GetPolicyDocumentStatementPrincipalArgs(
                            type="Service", identifiers=["logging.s3.amazonaws.com"]
                        )
                    ],
                    actions=["s3:GetBucketAcl", "s3:GetBucketPolicy"],
                    resources=[f"arn:aws:s3:::{bucket}"],
                    conditions=[
                        iam.GetPolicyDocumentStatementConditionArgs(
                            test="StringEquals",
                            variable="aws:SourceAccount",
                            values=[account_id],
                        ),
                        # iam.GetPolicyDocumentStatementConditionArgs(
                        #     test="ArnLike",
                        #     variable="AWS:SourceArn",
                        #     values=[f"arn:aws:logs:*:{account_id}:*"],
                        # ),
                    ],
                )
            )

            statements.append(
                iam.GetPolicyDocumentStatementArgs(
                    sid=f"S3ServerAccessLogsPolicy_{account_id}",
                    effect="Allow",
                    principals=[
                        iam.GetPolicyDocumentStatementPrincipalArgs(
                            identifiers=["logging.s3.amazonaws.com"], type="Service"
                        )
                    ],
                    actions=["s3:PutObject"],
                    resources=[f"arn:aws:s3:::{bucket}/*"],
                    conditions=[
                        iam.GetPolicyDocumentStatementConditionArgs(
                            test="ArnLike",
                            variable="aws:SourceArn",
                            values=[f"arn:aws:s3:::{bucket}"],
                        ),
                        iam.GetPolicyDocumentStatementConditionArgs(
                            test="StringEquals",
                            variable="aws:SourceAccount",
                            values=[account_id],
                        ),
                    ],
                )
            )

        return super().create(bucket, extra_statements=statements)


class BucketFactory(BaseFactory):
    def __init__(
        self,
        name: str,
        namespace: str = "epsy",
        opts: Optional[pulumi.ResourceOptions] = None,
        acceleration_status: Optional[pulumi.Input[str]] = None,
        acl: Optional[pulumi.Input[Union[str, "CannedAcl"]]] = None,
        arn: Optional[pulumi.Input[str]] = None,
        bucket: Optional[pulumi.Input[str]] = None,
        bucket_prefix: Optional[pulumi.Input[str]] = None,
        cors_rules: Optional[pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["BucketCorsRuleArgs"]]]]] = None,
        force_destroy: Optional[pulumi.Input[bool]] = None,
        grants: Optional[pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["BucketGrantArgs"]]]]] = None,
        hosted_zone_id: Optional[pulumi.Input[str]] = None,
        lifecycle_rules: Optional[
            pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["BucketLifecycleRuleArgs"]]]]
        ] = None,
        loggings: Optional[pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["BucketLoggingArgs"]]]]] = None,
        object_lock_configuration: Optional[pulumi.Input[pulumi.InputType["BucketObjectLockConfigurationArgs"]]] = None,
        policy: Optional[pulumi.Input[str]] = None,
        replication_configuration: Optional[
            pulumi.Input[pulumi.InputType["BucketReplicationConfigurationArgs"]]
        ] = None,
        request_payer: Optional[pulumi.Input[str]] = None,
        server_side_encryption_configuration: Optional[
            pulumi.Input[pulumi.InputType["BucketServerSideEncryptionConfigurationArgs"]]
        ] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        versioning: Optional[pulumi.Input[pulumi.InputType["BucketVersioningArgs"]]] = None,
        website: Optional[pulumi.Input[pulumi.InputType["BucketWebsiteArgs"]]] = None,
        website_domain: Optional[pulumi.Input[str]] = None,
        website_endpoint: Optional[pulumi.Input[str]] = None,
    ):
        super().__init__(
            resource_name=name,
            opts=opts,
            acceleration_status=acceleration_status,
            acl=acl,
            arn=arn,
            bucket=bucket,
            bucket_prefix=bucket_prefix,
            cors_rules=cors_rules,
            force_destroy=force_destroy,
            grants=grants,
            hosted_zone_id=hosted_zone_id,
            lifecycle_rules=lifecycle_rules,
            loggings=loggings,
            object_lock_configuration=object_lock_configuration,
            policy=policy,
            replication_configuration=replication_configuration,
            request_payer=request_payer,
            server_side_encryption_configuration=server_side_encryption_configuration,
            tags=tags,
            versioning=versioning,
            website=website,
            website_domain=website_domain,
            website_endpoint=website_endpoint,
        )
        # self.name = name
        self.policy = None
        self.namespace = namespace
        self.full_name = regional_domain(self.namespace, sub=name)
        self._enable_acl = False

    def with_policy(self, policy):
        self.policy = policy

    def enable_acl(self):
        self._enable_acl = True

    def create(self) -> s3classic.Bucket:
        identity = get_caller_identity()
        region = get_region()

        # self.set_default("access_control", s3.BucketAccessControl.PRIVATE)
        self.setdefault("acl", "private")
        self.setdefault("versioning", s3classic.BucketVersioningArgs(enabled=True))
        self.setdefault(
            "server_side_encryption_configuration",
            s3classic.BucketServerSideEncryptionConfigurationArgs(
                rule=s3classic.BucketServerSideEncryptionConfigurationRuleArgs(
                    apply_server_side_encryption_by_default=s3classic.BucketServerSideEncryptionConfigurationRuleApplyServerSideEncryptionByDefaultArgs(
                        sse_algorithm="aws:kms",
                        kms_master_key_id=f"arn:aws:kms:{region.name}:{identity.account_id}:alias/s3",
                    ),
                    bucket_key_enabled=True,
                )
            ),
        )

        self.setdefault(
            "object_lock_configuration",
            s3classic.BucketObjectLockConfigurationArgs(
                object_lock_enabled="Enabled",
            ),
        )

        self.setdefault("bucket", self.full_name)
        self.setdefault("opts", ResourceOptions(replace_on_changes=["bucket"], retain_on_delete=True))
        self.setdefault(
            "lifecycle_rules",
            [
                s3classic.BucketLifecycleRuleArgs(
                    enabled=True,
                    id="archive",
                    noncurrent_version_expiration=s3classic.BucketLifecycleRuleNoncurrentVersionExpirationArgs(days=90),
                    noncurrent_version_transitions=[
                        s3classic.BucketLifecycleRuleNoncurrentVersionTransitionArgs(
                            days=30, storage_class="STANDARD_IA"
                        ),
                        s3classic.BucketLifecycleRuleNoncurrentVersionTransitionArgs(days=60, storage_class="GLACIER"),
                    ],
                )
            ],
        )

        bucket = self._create(s3classic.Bucket)

        if self._enable_acl:
            s3classic.BucketOwnershipControls(
                self.identifier,
                bucket=bucket.bucket,
                rule=s3classic.BucketOwnershipControlsRuleArgs(
                    object_ownership="BucketOwnerPreferred",
                ),
            )

        policy = None
        if self.policy:
            if isinstance(self.policy, BucketPolicyFactory):
                policy = self.policy.create(self.full_name)
            else:
                policy = self.policy

        if not policy:
            policy = BucketPolicyFactory().create(self.full_name)

        s3classic.BucketPolicy(IdentifierV2(self.full_name + "-policy").snake, bucket=bucket.id, policy=policy.json)

        s3classic.BucketNotification(self.identifier.snake, bucket=bucket.id, eventbridge=True)

        return bucket
