from typing import Optional, Mapping, Sequence

import pulumi
from netaddr import IPNetwork
from pulumi_aws import ec2, get_caller_identity

from cloudlib.modules.ec2 import SecurityGroup
from cloudlib.modules.resource import BaseResourceMixin
from cloudlib.modules.vpc import transitgateway
from cloudlib.utils.names import IdentifierV2


class DefaultRouteTable(ec2.DefaultRouteTable, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        default_route_table_id: Optional[pulumi.Input[str]] = None,
        propagating_vgws: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        routes: Optional[pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["DefaultRouteTableRouteArgs"]]]]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        __props__=None,
        **kwargs,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=ec2.DefaultRouteTable,
            resource_name=name,
            opts=opts,
            default_route_table_id=default_route_table_id,
            propagating_vgws=propagating_vgws,
            routes=routes,
            tags=tags,
            __props__=__props__,
            **kwargs,
        )


class InternetGateway(ec2.InternetGateway, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        vpc_id: Optional[pulumi.Input[str]] = None,
        __props__=None,
        **kwargs,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=ec2.InternetGateway,
            resource_name=name,
            opts=opts,
            tags=tags,
            vpc_id=vpc_id,
            __props__=__props__,
            **kwargs,
        )


class NatGateway(ec2.NatGateway, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        allocation_id: Optional[pulumi.Input[str]] = None,
        connectivity_type: Optional[pulumi.Input[str]] = None,
        private_ip: Optional[pulumi.Input[str]] = None,
        secondary_allocation_ids: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        secondary_private_ip_address_count: Optional[pulumi.Input[int]] = None,
        secondary_private_ip_addresses: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        subnet_id: Optional[pulumi.Input[str]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        __props__=None,
        **kwargs,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=ec2.NatGateway,
            resource_name=name,
            opts=opts,
            allocation_id=allocation_id,
            connectivity_type=connectivity_type,
            private_ip=private_ip,
            secondary_allocation_ids=secondary_allocation_ids,
            secondary_private_ip_address_count=secondary_private_ip_address_count,
            secondary_private_ip_addresses=secondary_private_ip_addresses,
            subnet_id=subnet_id,
            tags=tags,
            __props__=__props__,
            **kwargs,
        )


class Eip(ec2.Eip, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        address: Optional[pulumi.Input[str]] = None,
        associate_with_private_ip: Optional[pulumi.Input[str]] = None,
        customer_owned_ipv4_pool: Optional[pulumi.Input[str]] = None,
        domain: Optional[pulumi.Input[str]] = None,
        instance: Optional[pulumi.Input[str]] = None,
        network_border_group: Optional[pulumi.Input[str]] = None,
        network_interface: Optional[pulumi.Input[str]] = None,
        public_ipv4_pool: Optional[pulumi.Input[str]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        vpc: Optional[pulumi.Input[bool]] = None,
        __props__=None,
        **kwargs,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=ec2.Eip,
            resource_name=name,
            opts=opts,
            address=address,
            associate_with_private_ip=associate_with_private_ip,
            customer_owned_ipv4_pool=customer_owned_ipv4_pool,
            domain=domain,
            instance=instance,
            network_border_group=network_border_group,
            network_interface=network_interface,
            public_ipv4_pool=public_ipv4_pool,
            tags=tags,
            vpc=vpc,
            __props__=__props__,
            **kwargs,
        )


class Subnet(ec2.Subnet, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        assign_ipv6_address_on_creation: Optional[pulumi.Input[bool]] = None,
        availability_zone: Optional[pulumi.Input[str]] = None,
        availability_zone_id: Optional[pulumi.Input[str]] = None,
        cidr_block: Optional[pulumi.Input[str]] = None,
        customer_owned_ipv4_pool: Optional[pulumi.Input[str]] = None,
        enable_dns64: Optional[pulumi.Input[bool]] = None,
        enable_lni_at_device_index: Optional[pulumi.Input[int]] = None,
        enable_resource_name_dns_a_record_on_launch: Optional[pulumi.Input[bool]] = None,
        enable_resource_name_dns_aaaa_record_on_launch: Optional[pulumi.Input[bool]] = None,
        ipv6_cidr_block: Optional[pulumi.Input[str]] = None,
        ipv6_native: Optional[pulumi.Input[bool]] = None,
        map_customer_owned_ip_on_launch: Optional[pulumi.Input[bool]] = None,
        map_public_ip_on_launch: Optional[pulumi.Input[bool]] = None,
        outpost_arn: Optional[pulumi.Input[str]] = None,
        private_dns_hostname_type_on_launch: Optional[pulumi.Input[str]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        vpc_id: Optional[pulumi.Input[str]] = None,
        __props__=None,
        **kwargs,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=ec2.Subnet,
            resource_name=name,
            opts=opts,
            assign_ipv6_address_on_creation=assign_ipv6_address_on_creation,
            availability_zone=availability_zone,
            availability_zone_id=availability_zone_id,
            cidr_block=cidr_block,
            customer_owned_ipv4_pool=customer_owned_ipv4_pool,
            enable_dns64=enable_dns64,
            enable_lni_at_device_index=enable_lni_at_device_index,
            enable_resource_name_dns_a_record_on_launch=enable_resource_name_dns_a_record_on_launch,
            enable_resource_name_dns_aaaa_record_on_launch=enable_resource_name_dns_aaaa_record_on_launch,
            ipv6_cidr_block=ipv6_cidr_block,
            ipv6_native=ipv6_native,
            map_customer_owned_ip_on_launch=map_customer_owned_ip_on_launch,
            map_public_ip_on_launch=map_public_ip_on_launch,
            outpost_arn=outpost_arn,
            private_dns_hostname_type_on_launch=private_dns_hostname_type_on_launch,
            tags=tags,
            vpc_id=vpc_id,
            __props__=__props__,
            **kwargs,
        )

    def nat_gateway(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        allocation_id: Optional[pulumi.Input[str]] = None,
        connectivity_type: Optional[pulumi.Input[str]] = None,
        private_ip: Optional[pulumi.Input[str]] = None,
        secondary_allocation_ids: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        secondary_private_ip_address_count: Optional[pulumi.Input[int]] = None,
        secondary_private_ip_addresses: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        subnet_id: Optional[pulumi.Input[str]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        __props__=None,
        **kwargs,
    ):
        if not allocation_id:
            eip = Eip(name=self.identifier.add("nat-ip"), args=ec2.EipArgs())
            allocation_id = eip.allocation_id

        if not subnet_id:
            subnet_id = self.id

        return NatGateway(
            name=self.identifier.add(name or "nat-gw"),
            opts=opts,
            allocation_id=allocation_id,
            connectivity_type=connectivity_type,
            private_ip=private_ip,
            secondary_allocation_ids=secondary_allocation_ids,
            secondary_private_ip_address_count=secondary_private_ip_address_count,
            secondary_private_ip_addresses=secondary_private_ip_addresses,
            subnet_id=subnet_id,
            tags=tags,
            __props__=__props__,
            **kwargs,
        )


class RouteTableAssociation(ec2.RouteTableAssociation, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        gateway_id: Optional[pulumi.Input[str]] = None,
        route_table_id: Optional[pulumi.Input[str]] = None,
        subnet_id: Optional[pulumi.Input[str]] = None,
        __props__=None,
        **kwargs,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=ec2.RouteTableAssociation,
            resource_name=name,
            opts=opts,
            gateway_id=gateway_id,
            route_table_id=route_table_id,
            subnet_id=subnet_id,
            __props__=__props__,
            **kwargs,
        )


class Route(ec2.Route):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        carrier_gateway_id: Optional[pulumi.Input[str]] = None,
        core_network_arn: Optional[pulumi.Input[str]] = None,
        destination_cidr_block: Optional[pulumi.Input[str]] = None,
        destination_ipv6_cidr_block: Optional[pulumi.Input[str]] = None,
        destination_prefix_list_id: Optional[pulumi.Input[str]] = None,
        egress_only_gateway_id: Optional[pulumi.Input[str]] = None,
        gateway_id: Optional[pulumi.Input[str]] = None,
        local_gateway_id: Optional[pulumi.Input[str]] = None,
        nat_gateway_id: Optional[pulumi.Input[str]] = None,
        network_interface_id: Optional[pulumi.Input[str]] = None,
        route_table_id: Optional[pulumi.Input[str]] = None,
        transit_gateway_id: Optional[pulumi.Input[str]] = None,
        vpc_endpoint_id: Optional[pulumi.Input[str]] = None,
        vpc_peering_connection_id: Optional[pulumi.Input[str]] = None,
        __props__=None,
        **kwargs,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=ec2.Route,
            resource_name=name,
            opts=opts,
            carrier_gateway_id=carrier_gateway_id,
            core_network_arn=core_network_arn,
            destination_cidr_block=destination_cidr_block,
            destination_ipv6_cidr_block=destination_ipv6_cidr_block,
            destination_prefix_list_id=destination_prefix_list_id,
            egress_only_gateway_id=egress_only_gateway_id,
            gateway_id=gateway_id,
            local_gateway_id=local_gateway_id,
            nat_gateway_id=nat_gateway_id,
            network_interface_id=network_interface_id,
            route_table_id=route_table_id,
            transit_gateway_id=transit_gateway_id,
            vpc_endpoint_id=vpc_endpoint_id,
            vpc_peering_connection_id=vpc_peering_connection_id,
            __props__=__props__,
            **kwargs,
        )


class RouteTable(ec2.RouteTable, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        propagating_vgws: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        routes: Optional[pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["RouteTableRouteArgs"]]]]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        vpc_id: Optional[pulumi.Input[str]] = None,
        associate_with=None,
        __props__=None,
        **kwargs,
    ):
        BaseResourceMixin.__init__(
            self,
            resource_name=name,
            parent=ec2.RouteTable,
            opts=opts,
            propagating_vgws=propagating_vgws,
            routes=routes,
            tags=tags,
            vpc_id=vpc_id,
            __props__=__props__,
            **kwargs,
        )

        if associate_with:
            RouteTableAssociation(
                name=self.identifier.add("association"),
                route_table_id=self.id,
                **associate_with,
            )

    def add_route(
        self,
        name,
        carrier_gateway_id: Optional[pulumi.Input[str]] = None,
        core_network_arn: Optional[pulumi.Input[str]] = None,
        destination_cidr_block: Optional[pulumi.Input[str]] = None,
        destination_ipv6_cidr_block: Optional[pulumi.Input[str]] = None,
        destination_prefix_list_id: Optional[pulumi.Input[str]] = None,
        egress_only_gateway_id: Optional[pulumi.Input[str]] = None,
        gateway_id: Optional[pulumi.Input[str]] = None,
        local_gateway_id: Optional[pulumi.Input[str]] = None,
        nat_gateway_id: Optional[pulumi.Input[str]] = None,
        network_interface_id: Optional[pulumi.Input[str]] = None,
        transit_gateway_id: Optional[pulumi.Input[str]] = None,
        vpc_endpoint_id: Optional[pulumi.Input[str]] = None,
        vpc_peering_connection_id: Optional[pulumi.Input[str]] = None,
    ):
        return Route(
            name=name,
            route_table_id=self.id,
            carrier_gateway_id=carrier_gateway_id,
            core_network_arn=core_network_arn,
            destination_cidr_block=str(destination_cidr_block),
            destination_ipv6_cidr_block=destination_ipv6_cidr_block,
            destination_prefix_list_id=destination_prefix_list_id,
            egress_only_gateway_id=egress_only_gateway_id,
            gateway_id=gateway_id,
            local_gateway_id=local_gateway_id,
            nat_gateway_id=nat_gateway_id,
            network_interface_id=network_interface_id,
            transit_gateway_id=transit_gateway_id,
            vpc_endpoint_id=vpc_endpoint_id,
            vpc_peering_connection_id=vpc_peering_connection_id,
        )


class VpcIpv4CidrBlockAssociation(ec2.VpcIpv4CidrBlockAssociation, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        cidr_block: Optional[pulumi.Input[str]] = None,
        ipv4_ipam_pool_id: Optional[pulumi.Input[str]] = None,
        ipv4_netmask_length: Optional[pulumi.Input[int]] = None,
        vpc_id: Optional[pulumi.Input[str]] = None,
        __props__=None,
        **kwargs,
    ):
        BaseResourceMixin.__init__(
            self,
            resource_name=name,
            parent=ec2.VpcIpv4CidrBlockAssociation,
            opts=opts,
            cidr_block=cidr_block,
            ipv4_ipam_pool_id=ipv4_ipam_pool_id,
            ipv4_netmask_length=ipv4_netmask_length,
            vpc_id=vpc_id,
            __props__=__props__,
            **kwargs,
        )


class Vpc(ec2.Vpc, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        az_ids,
        parent=ec2.Vpc,
        subnet_size=19,
        opts: Optional[pulumi.ResourceOptions] = None,
        assign_generated_ipv6_cidr_block: Optional[pulumi.Input[bool]] = None,
        cidr_block: Optional[pulumi.Input[str]] = None,
        enable_dns_hostnames: Optional[pulumi.Input[bool]] = True,
        enable_dns_support: Optional[pulumi.Input[bool]] = True,
        enable_network_address_usage_metrics: Optional[pulumi.Input[bool]] = None,
        instance_tenancy: Optional[pulumi.Input[str]] = None,
        ipv4_ipam_pool_id: Optional[pulumi.Input[str]] = None,
        ipv4_netmask_length: Optional[pulumi.Input[int]] = None,
        ipv6_cidr_block: Optional[pulumi.Input[str]] = None,
        ipv6_cidr_block_network_border_group: Optional[pulumi.Input[str]] = None,
        ipv6_ipam_pool_id: Optional[pulumi.Input[str]] = None,
        ipv6_netmask_length: Optional[pulumi.Input[int]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        __props__=None,
        **kwargs,
    ):
        self.az_ids = az_ids
        self.subnet_size = subnet_size
        self._subnets = list(IPNetwork(cidr_block).subnet(self.subnet_size))
        self.networks = [cidr_block]
        self.network = IPNetwork(cidr_block)

        BaseResourceMixin.__init__(
            self,
            parent=ec2.Vpc,
            resource_name=name,
            opts=opts,
            assign_generated_ipv6_cidr_block=assign_generated_ipv6_cidr_block,
            cidr_block=str(cidr_block),
            enable_dns_hostnames=enable_dns_hostnames,
            enable_dns_support=enable_dns_support,
            enable_network_address_usage_metrics=enable_network_address_usage_metrics,
            instance_tenancy=instance_tenancy,
            ipv4_ipam_pool_id=ipv4_ipam_pool_id,
            ipv4_netmask_length=ipv4_netmask_length,
            ipv6_cidr_block=ipv6_cidr_block,
            ipv6_cidr_block_network_border_group=ipv6_cidr_block_network_border_group,
            ipv6_ipam_pool_id=ipv6_ipam_pool_id,
            ipv6_netmask_length=ipv6_netmask_length,
            tags=tags,
            __props__=__props__,
            **kwargs,
        )

    def add_cidr(self, cidr):
        self.networks.append(cidr)
        VpcIpv4CidrBlockAssociation(
            self.identifier.add(f"network {len(self.networks)}"), vpc_id=self.id, cidr_block=str(cidr)
        )

    def internet_gateway(self, name: str = None, **kwargs):
        kwargs.setdefault("vpc_id", self.id)

        return InternetGateway(name=self.identifier.add(name or "internet-gw"), **kwargs)

    def subnets(self, prefix, with_routing_table=False, **kwargs):
        subnet_cidr: IPNetwork = self._subnets.pop(0)
        zone_subnet = list(subnet_cidr.subnet(self.subnet_size + 1))
        subnets = {}
        for idx, zone in enumerate(self.az_ids):
            local_kwargs = kwargs.copy()
            local_kwargs["availability_zone_id"] = zone
            local_kwargs["cidr_block"] = str(zone_subnet[idx])

            subnet = Subnet(
                name=self.identifier.add(prefix).add(zone),
                vpc_id=self.id,
                tags={"Name": IdentifierV2(prefix).add(zone).spinal, "Type": prefix.lower()},
                **local_kwargs,
            )
            subnet._zone_id = zone

            if with_routing_table:
                subnet.new_routing_table = RouteTable(
                    name=subnet.identifier.add("routing"),
                    vpc_id=self.id,
                    associate_with=dict(subnet_id=subnet.id),
                )

            subnets[zone] = subnet

        return subnets

    def security_group(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        description: Optional[pulumi.Input[str]] = None,
        egress: Optional[pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["SecurityGroupEgressArgs"]]]]] = None,
        ingress: Optional[pulumi.Input[Sequence[pulumi.Input[pulumi.InputType["SecurityGroupIngressArgs"]]]]] = None,
        name_prefix: Optional[pulumi.Input[str]] = None,
        revoke_rules_on_delete: Optional[pulumi.Input[bool]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        __props__=None,
    ):
        return SecurityGroup(
            self.identifier.add(name),
            opts=opts,
            description=description,
            egress=egress,
            ingress=ingress,
            name_prefix=name_prefix,
            revoke_rules_on_delete=revoke_rules_on_delete,
            tags=tags,
            vpc_id=self.id,
        )

    def firewall_subnets(self, firewall_network, transit_gateway=None):
        firewall_subnets = {}
        for zone, subnet_cidr in dict(zip(self.az_ids, list(firewall_network.subnet(28)))).items():
            subnet = Subnet(
                name=self.identifier.add("firewall").add(zone),
                cidr_block=str(subnet_cidr),
                vpc_id=self.id,
                availability_zone_id=zone,
            )

            rt = RouteTable(
                name=subnet.identifier.add("routing"),
                vpc_id=self.id,
                associate_with=dict(subnet_id=subnet.id),
            )

            subnet.new_routing_table = rt

            firewall_subnets[zone] = subnet
        return firewall_subnets

    def transit_subnets(
        self,
        transit_network,
        transit_gateway=None,
        region_cidr=None,
        create_routing_table=True,
        include_region_route=True,
    ):
        transit_subnets = {}
        for zone, subnet_cidr in dict(zip(self.az_ids, list(transit_network.subnet(28)))).items():
            subnet = Subnet(
                name=self.identifier.add("transit").add(zone),
                cidr_block=str(subnet_cidr),
                vpc_id=self.id,
                availability_zone_id=zone,
            )

            transit_subnets[zone] = subnet
            if not create_routing_table:
                continue

            rt = RouteTable(
                name=subnet.identifier.add("routing"),
                vpc_id=self.id,
                associate_with=dict(subnet_id=subnet.id),
            )
            subnet.new_routing_table = rt

            if include_region_route:
                rt.add_route(
                    name=rt.identifier.add("region"),
                    destination_cidr_block=str(region_cidr),
                    transit_gateway_id=transit_gateway.id,
                )

        return transit_subnets

    def attach_to(
        self,
        transit_gateway,
        subnet_ids,
        create_custom_route_table=False,
        propagate=False,
        route_table_default_propagation_route_table=False,
        **kwargs,
    ):
        kwargs.setdefault("transit_gateway_default_route_table_association", not create_custom_route_table)
        kwargs.setdefault("transit_gateway_default_route_table_propagation", not create_custom_route_table)
        attachment = transitgateway.VpcAttachment(
            name=self.identifier.add("attachment").add(transit_gateway.identifier),
            subnet_ids=subnet_ids,
            transit_gateway_id=transit_gateway.id,
            vpc_id=self.id,
            appliance_mode_support="enable",
            **kwargs,
        )

        if create_custom_route_table:
            table = attachment.create_route_table()
            association = transitgateway.RouteTableAssociation(
                table.identifier.add("association"),
                transit_gateway_route_table_id=table.id,
                transit_gateway_attachment_id=attachment.id,
            )

            if propagate:
                association.propagate()

            attachment.route_table = table

        return attachment

    def with_flow_logs(self, bucket):
        identity = get_caller_identity()
        ec2.FlowLog(
            self.identifier,
            log_destination=bucket,
            log_destination_type="s3",
            log_format="${version} ${account-id} ${interface-id} ${srcaddr} ${dstaddr} ${srcport} ${dstport} ${protocol} ${packets} ${bytes} ${start} ${end} ${action} ${log-status} ${vpc-id} ${subnet-id} ${instance-id} ${tcp-flags} ${type} ${pkt-srcaddr} ${pkt-dstaddr} ${region} ${az-id} ${sublocation-type} ${sublocation-id} ${pkt-src-aws-service} ${pkt-dst-aws-service} ${flow-direction} ${traffic-path}",
            traffic_type="ALL",
            vpc_id=self.id,
            destination_options=ec2.FlowLogDestinationOptionsArgs(
                file_format="parquet",
                per_hour_partition=True,
            ),
        )
