from typing import Optional, Sequence, Mapping, overload

import pulumi
from pulumi_aws import ec2transitgateway

from cloudlib.modules.ram import ResourceShare, ResourceAssociation, PrincipalAssociation
from cloudlib.modules.resource import BaseResourceMixin
from cloudlib.utils.names import IdentifierV2
from pulumi_aws.ec2transitgateway import RouteTablePropagationArgs


class TransitGateway(ec2transitgateway.TransitGateway, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        amazon_side_asn: Optional[pulumi.Input[int]] = None,
        auto_accept_shared_attachments: Optional[pulumi.Input[str]] = None,
        default_route_table_association: Optional[pulumi.Input[str]] = None,
        default_route_table_propagation: Optional[pulumi.Input[str]] = None,
        description: Optional[pulumi.Input[str]] = None,
        dns_support: Optional[pulumi.Input[str]] = None,
        multicast_support: Optional[pulumi.Input[str]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        transit_gateway_cidr_blocks: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        vpn_ecmp_support: Optional[pulumi.Input[str]] = None,
        __props__=None,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=ec2transitgateway.TransitGateway,
            resource_name=name,
            opts=opts,
            amazon_side_asn=amazon_side_asn,
            auto_accept_shared_attachments=auto_accept_shared_attachments,
            default_route_table_association=default_route_table_association,
            default_route_table_propagation=default_route_table_propagation,
            description=description,
            dns_support=dns_support,
            multicast_support=multicast_support,
            tags=tags,
            transit_gateway_cidr_blocks=transit_gateway_cidr_blocks,
            vpn_ecmp_support=vpn_ecmp_support,
            __props__=__props__,
        )

    def share(self, accounts):
        share = ResourceShare(name=self.identifier.add("share"), allow_external_principals=False)
        association = ResourceAssociation(
            name=share.identifier.add("transitgateway"), resource_arn=self.arn, resource_share_arn=share.arn
        )
        for account in accounts or []:
            PrincipalAssociation(
                name=share.identifier.add(f"share-{account}"), principal=str(account), resource_share_arn=share.arn
            )

        return share


class VpcAttachment(ec2transitgateway.VpcAttachment, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        appliance_mode_support: Optional[pulumi.Input[str]] = None,
        dns_support: Optional[pulumi.Input[str]] = None,
        ipv6_support: Optional[pulumi.Input[str]] = None,
        subnet_ids: Optional[pulumi.Input[Sequence[pulumi.Input[str]]]] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        transit_gateway_default_route_table_association: Optional[pulumi.Input[bool]] = None,
        transit_gateway_default_route_table_propagation: Optional[pulumi.Input[bool]] = None,
        transit_gateway_id: Optional[pulumi.Input[str]] = None,
        vpc_id: Optional[pulumi.Input[str]] = None,
        __props__=None,
        **kwargs,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=ec2transitgateway.VpcAttachment,
            resource_name=name,
            opts=opts,
            appliance_mode_support=appliance_mode_support,
            dns_support=dns_support,
            ipv6_support=ipv6_support,
            subnet_ids=subnet_ids,
            tags=tags,
            transit_gateway_default_route_table_association=transit_gateway_default_route_table_association,
            transit_gateway_default_route_table_propagation=transit_gateway_default_route_table_propagation,
            transit_gateway_id=transit_gateway_id,
            vpc_id=vpc_id,
            __props__=__props__,
            **kwargs,
        )

    def create_route_table(self):
        return RouteTable(name=self.identifier.add("route-table"), transit_gateway_id=self.transit_gateway_id)


class RouteTableAssociation(ec2transitgateway.RouteTableAssociation, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        replace_existing_association: Optional[pulumi.Input[bool]] = None,
        transit_gateway_attachment_id: Optional[pulumi.Input[str]] = None,
        transit_gateway_route_table_id: Optional[pulumi.Input[str]] = None,
        __props__=None,
    ):
        BaseResourceMixin.__init__(
            self,
            resource_name=name,
            parent=ec2transitgateway.RouteTableAssociation,
            opts=opts,
            replace_existing_association=replace_existing_association,
            transit_gateway_route_table_id=transit_gateway_route_table_id,
            transit_gateway_attachment_id=transit_gateway_attachment_id,
            __props__=__props__,
        )

    def propagate(self):
        RouteTablePropagation(
            name=self.identifier.add("propagation"),
            transit_gateway_route_table_id=self.transit_gateway_route_table_id,
            transit_gateway_attachment_id=self.transit_gateway_attachment_id,
        )


class RouteTablePropagation(ec2transitgateway.RouteTablePropagation, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        transit_gateway_attachment_id: Optional[pulumi.Input[str]] = None,
        transit_gateway_route_table_id: Optional[pulumi.Input[str]] = None,
        __props__=None,
        **kwargs,
    ):
        BaseResourceMixin.__init__(
            self,
            resource_name=name,
            parent=ec2transitgateway.RouteTablePropagation,
            opts=opts,
            transit_gateway_route_table_id=transit_gateway_route_table_id,
            transit_gateway_attachment_id=transit_gateway_attachment_id,
            __props__=__props__,
            **kwargs,
        )


class RouteTable(ec2transitgateway.RouteTable, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        tags: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] = None,
        transit_gateway_id: Optional[pulumi.Input[str]] = None,
        __props__=None,
        **kwargs,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=ec2transitgateway.RouteTable,
            resource_name=name,
            opts=opts,
            tags=tags,
            transit_gateway_id=transit_gateway_id,
            __props__=__props__,
            **kwargs,
        )

    def add_route(
        self,
        name,
        opts: Optional[pulumi.ResourceOptions] = None,
        blackhole: Optional[pulumi.Input[bool]] = None,
        destination_cidr_block: Optional[pulumi.Input[str]] = None,
        transit_gateway_attachment_id: Optional[pulumi.Input[str]] = None,
        transit_gateway_route_table_id: Optional[pulumi.Input[str]] = None,
        __props__=None,
        **kwargs,
    ):
        transit_gateway_route_table_id = transit_gateway_route_table_id or self.id

        return Route(
            name=name,
            opts=opts,
            blackhole=blackhole,
            destination_cidr_block=destination_cidr_block,
            transit_gateway_route_table_id=transit_gateway_route_table_id,
            transit_gateway_attachment_id=transit_gateway_attachment_id,
            __props__=__props__,
            **kwargs,
        )


class Route(ec2transitgateway.Route, BaseResourceMixin):
    def __init__(
        self,
        name: str,
        opts: Optional[pulumi.ResourceOptions] = None,
        blackhole: Optional[pulumi.Input[bool]] = None,
        destination_cidr_block: Optional[pulumi.Input[str]] = None,
        transit_gateway_attachment_id: Optional[pulumi.Input[str]] = None,
        transit_gateway_route_table_id: Optional[pulumi.Input[str]] = None,
        __props__=None,
        **kwargs,
    ):
        BaseResourceMixin.__init__(
            self,
            parent=ec2transitgateway.Route,
            resource_name=name,
            opts=opts,
            blackhole=blackhole,
            destination_cidr_block=destination_cidr_block,
            transit_gateway_route_table_id=transit_gateway_route_table_id,
            transit_gateway_attachment_id=transit_gateway_attachment_id,
            __props__=__props__,
            **kwargs,
        )
