import pulumi

from cloudlib.aws.cloudfront import CloudFrontBucketAccessLogPolicyFactory
from cloudlib.aws.ssm import parameter_plain
from cloudlib.modules.cloudtrail import TrailFactory
from cloudlib.modules.kms import KeyFactory
from cloudlib.modules.s3 import BucketFactory, S3BucketAccessLogPolicyFactory
from cloudlib.shared.security import (
    account_security_forwarding,
    account_security,
    account_security_hub,
    roles,
    ecr_security,
    waf,
    account_notifications,
)

from pulumi_aws import get_caller_identity

identity = get_caller_identity()


def security(
    app,
    config,
    stage,
    region,
    with_audit=True,
    with_roles=True,
    with_waf=True,
    with_s3_logs_bucket=True,
    with_cloudfront_logs_bucket=True,
    with_cloudtrail=True,
    with_global_waf=True,
    white_list_ips=None
):
    if with_audit:
        audit_stack = pulumi.StackReference(f"LivaNova/stage-audit/{stage}-{region}")

        account_security_forwarding(config, region, stage, audit_stack)
        account_security(config, region, stage)
        account_security_hub(config, region, stage)

        if with_waf:
            waf(audit_stack, name="api", scope="regional", white_list_ips=white_list_ips, disable_graphql_schema=True, allow_localhost=True)
            if with_global_waf:
                waf(audit_stack, name="cdn", scope="cloudfront", white_list_ips=white_list_ips, allow_localhost=True)

        if with_cloudtrail:
            trail_factory = TrailFactory(
                name=stage,
                s3_bucket_name=audit_stack.get_output("cloudtrail_bucket"),
                kms_key_id=audit_stack.get_output("cloudtrail_kms_key"),
            )
            trail_factory.with_cloudwatch("cloudtrail/cloud-audit")

            account_notifications(config, region, stage)

    if with_roles:
            roles(config, region, stage)

    ecr_security()

    secrets_factory = KeyFactory("secrets")
    secrets_factory.with_alias("secrets")
    secrets_factory.create()

    if with_s3_logs_bucket or with_cloudfront_logs_bucket:
        s3_key_factory = KeyFactory("s3")
        s3_key_factory.policy["Statement"].append(
            {
                "Effect": "Allow",
                "Principal": {"Service": ["logging.s3.amazonaws.com"]},
                "Action": "kms:GenerateDataKey*",
                "Resource": "*",
                "Condition": {"StringEquals": {"aws:SourceAccount": identity.account_id}},
            }
        )
        s3_key_factory.with_alias("s3")
        key = s3_key_factory.create()
        parameter_plain("/audit/s3/bucket/key", key.arn)

    if with_s3_logs_bucket:
        s3_logs_factory = BucketFactory(f"logs.s3.{app}")
        s3_logs_factory.with_policy(S3BucketAccessLogPolicyFactory())
        s3_logs_bucket = s3_logs_factory.create()
        parameter_plain("/audit/s3/bucket", s3_logs_bucket.bucket)
        pulumi.export("s3_logs_bucket", s3_logs_bucket.bucket)

    if with_cloudfront_logs_bucket:
        cloudfront_logs_factory = BucketFactory(f"logs.cloudfront.{app}")
        cloudfront_logs_factory.with_policy(CloudFrontBucketAccessLogPolicyFactory())
        cloudfront_logs_factory.enable_acl()
        cloudfront_logs_bucket = cloudfront_logs_factory.create()
        parameter_plain("/audit/cloudfront/bucket", cloudfront_logs_bucket.bucket_regional_domain_name)
        pulumi.export("cloudfront_logs_bucket", cloudfront_logs_bucket.bucket_domain_name)
