import json

import pulumi
import pulumi_aws
from pulumi_aws import provider, organizations, iam

from cloudlib.aws.ssm import parameter_plain
from cloudlib.modules.iam import RoleFactory
from cloudlib.modules.route53 import ZoneFactory, ResolverRuleAssociation, ZoneAssociation, VpcAssociationAuthorization
from cloudlib.utils.names import IdentifierV2


def iam_dns_role(config, zones):
    organization_provider = provider.Provider(
        resource_name=IdentifierV2("provider_organization").snake,
        assume_role=provider.ProviderAssumeRoleArgs(
            external_id="078020966649",
            role_arn="arn:aws:iam::078020966649:role/OrganizationInfoRole",
        ),
    )

    org_unit = organizations.get_organizational_unit_descendant_accounts(
        parent_id=config.get("organization-unit"),
        opts=pulumi.InvokeOptions(provider=organization_provider),
    )

    role_factory = RoleFactory(
        name="NMDomainManager",
        assume_role_policy=json.dumps(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Action": "sts:AssumeRole",
                        "Effect": "Allow",
                        "Sid": "",
                        "Principal": {"AWS": [account.id for account in org_unit.accounts]},
                    }
                ],
            }
        ),
        inline_policies=[
            iam.RoleInlinePolicyArgs(
                name="allow_zone_association",
                policy=pulumi.Output.all(zones[0], zones[1]).apply(
                    lambda izones: json.dumps(
                        {
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Action": [
                                        "route53:AssociateVPCWithHostedZone",
                                        "route53:ListHostedZonesByVPC",
                                        "route53:DisassociateVPCFromHostedZone",
                                    ],
                                    "Effect": "Allow",
                                    "Resource": "*",
                                },
                                {
                                    "Action": [
                                        "ec2:DescribeVpcs",
                                    ],
                                    "Effect": "Allow",
                                    "Resource": "*",
                                },
                                {
                                    "Action": [
                                        "route53:ListHostedZonesByVPC",
                                        "route53:ChangeResourceRecordSets",
                                        "route53:GetHostedZone",
                                    ],
                                    "Effect": "Allow",
                                    "Resource": [izones[0], izones[1]],
                                },
                                {
                                    "Action": ["route53:GetChange", "route53:ListResourceRecordSets"],
                                    "Effect": "Allow",
                                    "Resource": "*",
                                },
                            ],
                        }
                    ),
                ),
            )
        ],
    )

    role = role_factory.create()

    pulumi.export("dns_manager_role", role.arn)

    return role


def dns_configure(config, region_id, stage, vpc, core_networking, cloud, zone=True):
    zone_prefix = f"{cloud}.{region_id}.{stage}"

    if zone:
        nm_public_zone_factory = ZoneFactory(resource_name=f"{zone_prefix}.neuromodulation.cloud")
        nm_public_zone_factory.record(
            resource_name=f"{zone_prefix}_public_a",
            name="",
            type="A",
            ttl=600,
            records=["8.8.8.8"],
        )
        nm_public_zone = nm_public_zone_factory.create()
        nm_private_zone = ZoneFactory(
            resource_name=f"{zone_prefix}.neuromodulation.cloud", vpcs=[dict(vpc_id=vpc.id)]
        ).create()

        stage_dns_provider = pulumi_aws.provider.Provider(
            resource_name="stage_dns_provider",
            assume_role=pulumi_aws.provider.ProviderAssumeRoleArgs(
                role_arn=core_networking.get_output("dns_manager_role"),
            ),
        )

        zf = ZoneFactory(
            resource_name="private_zone", resource_id=core_networking.get_output("neuromodulation_private_zone_id")
        )
        zf.record(
            resource_name=f"{zone_prefix}_private_ns",
            name=f"{cloud}.{region_id}",
            type="NS",
            ttl=600,
            records=nm_private_zone.name_servers,
            opts=pulumi.ResourceOptions(provider=stage_dns_provider),
        )

        zf = ZoneFactory(
            resource_name="public_zone", resource_id=core_networking.get_output("neuromodulation_public_zone_id")
        )
        zf.record(
            resource_name=f"{zone_prefix}_public_ns",
            name=f"{cloud}.{region_id}",
            type="NS",
            ttl=600,
            records=nm_public_zone.name_servers,
            opts=pulumi.ResourceOptions(provider=stage_dns_provider),
        )

        role = iam_dns_role(config, zones=[nm_private_zone.arn, nm_public_zone.arn])

        zone_fqdn = f"{cloud}.{region_id}.{config.stage}.neuromodulation.cloud"

        association_authorization = VpcAssociationAuthorization(
            name="zone_share", vpc_id=core_networking.get_output("vpc_egress"), zone_id=nm_private_zone.zone_id
        )

        networking_provider = pulumi_aws.provider.Provider(
            resource_name="networking_provider",
            assume_role=pulumi_aws.provider.ProviderAssumeRoleArgs(
                role_arn=core_networking.get_output("dns_manager_role"),
            ),
        )

        ZoneAssociation(
            "zone_association",
            vpc_id=core_networking.get_output("vpc_egress"),
            zone_id=nm_private_zone.zone_id,
            opts=pulumi.ResourceOptions(provider=networking_provider),
        )

        pulumi.export("private_zone_id", nm_private_zone.zone_id)
        pulumi.export("private_zone_name", zone_fqdn)
        pulumi.export("public_zone_id", nm_public_zone.zone_id)
        pulumi.export("public_zone_name", zone_fqdn)

        parameter_plain(f"/route53/zones/app/private", nm_private_zone.zone_id)
        parameter_plain(f"/route53/domains/app/private", zone_fqdn)
        parameter_plain(f"/route53/zones/app/public", nm_public_zone.zone_id)
        parameter_plain(f"/route53/domains/app/public", zone_fqdn)

    core_networking.get_output("dns_rules").apply(
        lambda rules: [
            ResolverRuleAssociation(name=f"ingress_za_{name}", resolver_rule_id=rule, vpc_id=vpc.id)
            for name, rule in rules.items()
        ]
    )
