import pulumi, pulumi_aws
from pulumi import ResourceOptions
from pulumi_aws import ec2

from cloudlib.aws.ec2 import SecurityGroupFactory
from cloudlib.aws.ssm import parameter_plain
from cloudlib.cloudinfo import Region
from cloudlib.modules.route53 import Zone
from cloudlib.modules.vpc.network import Vpc, DefaultRouteTable
from cloudlib.utils.names import domain, regional_domain
from cloudlib.utils.resources import NetworkInfo


def domain_ssms(cloud, networking_stack):
    parameter_plain("/route53/domains/epsy/primary", domain())
    parameter_plain("/route53/domains/epsy/regional", regional_domain())
    parameter_plain("/route53/domains/livanova/primary", domain(scope="livanova"))
    parameter_plain("/route53/domains/livanova/regional", regional_domain(scope="livanova"))

    parameter_plain(
        "/route53/zones/livanova/primary", networking_stack.get_output("route53_zones")["livanova"]["zone_id"]
    )
    parameter_plain("/route53/zones/epsy/primary", networking_stack.get_output("route53_zones")["epsy"]["zone_id"])
    parameter_plain("/route53/management_role", networking_stack.get_output("dns_manager_role"))

    parameter_plain("/global/cloud-id", regional_domain(scope="epsy", cloud=cloud))


def default_acl(region_cidr, vpc, subnets):
    ec2.DefaultNetworkAcl(
        "default",
        default_network_acl_id=vpc.default_network_acl_id,
        subnet_ids=subnets,
        ingress=[
            ec2.DefaultNetworkAclIngressArgs(
                protocol="tcp",
                rule_no=100,
                action="allow",
                cidr_block=str(region_cidr),
                from_port=22,
                to_port=22,
            ),
            ec2.DefaultNetworkAclIngressArgs(
                protocol="tcp",
                rule_no=101,
                action="allow",
                cidr_block=str(region_cidr),
                from_port=3389,
                to_port=3389,
            ),
            ec2.DefaultNetworkAclIngressArgs(
                protocol="tcp",
                rule_no=110,
                action="deny",
                cidr_block="0.0.0.0/0",
                from_port=22,
                to_port=22,
            ),
            ec2.DefaultNetworkAclIngressArgs(
                protocol="tcp",
                rule_no=111,
                action="deny",
                cidr_block="0.0.0.0/0",
                from_port=3389,
                to_port=3389,
            ),
            ec2.DefaultNetworkAclIngressArgs(
                protocol="-1",
                rule_no=1000,
                action="allow",
                cidr_block="0.0.0.0/0",
                from_port=0,
                to_port=0,
            ),
        ],
        egress=[
            ec2.DefaultNetworkAclEgressArgs(
                protocol="-1",
                rule_no=100,
                action="allow",
                cidr_block="0.0.0.0/0",
                from_port=0,
                to_port=0,
            )
        ],
    )


def app_network(
    app_name, stage, region_name, transit_gateway, inspect_info: NetworkInfo = None, audit_stack=None, vpce=None
):
    region = Region.get(stage, region_name)
    app = region.network.get_app_network(app_name)

    vpc = Vpc(name=app_name, az_ids=app.az_ids, cidr_block=str(app.app_cidr), tags={"Name": "app"})

    vpc.add_cidr(app.shared_cidr)

    if audit_stack:
        vpc.with_flow_logs(audit_stack.get_output("flow_logs_bucket_arn"))

    ec2.DefaultSecurityGroup("default", vpc_id=vpc.id, ingress=[], egress=[])

    transit_subnets = vpc.transit_subnets(
        app.transit_network, transit_gateway, region.network.network_cidr, create_routing_table=False
    )

    attachment = vpc.attach_to(
        transit_gateway,
        [subnet.id for subnet in transit_subnets.values()],
        create_custom_route_table=False,
        propagate=False,
    )

    DefaultRouteTable(
        vpc.identifier.add("default-route-table"),
        default_route_table_id=vpc.default_route_table_id,
        routes=[{"cidr_block": "0.0.0.0/0", "transit_gateway_id": transit_gateway.id}],
        opts=ResourceOptions(depends_on=[attachment]),
    )

    subnets = vpc.subnets(prefix="private", with_routing_table=False)
    subnets_list = list(subnets.values())

    default_acl(region.network.network_cidr, vpc, subnets_list + list(transit_subnets.values()))

    parameter_plain("/vpc/subnet/private-1", subnets_list[0].id)
    parameter_plain("/vpc/subnet/private-2", subnets_list[1].id)

    zone = Zone(name=f"{app_name}.cloud.{stage}", vpcs=[dict(vpc_id=vpc.id)])

    if vpce:
        vpce_sg_factory = SecurityGroupFactory(name="vpce", vpc_id=vpc.id)
        vpce_sg_factory.with_https_egress()
        vpce_sg_factory.with_ingress(
            from_port=443, protocol="TCP", to_port=443, cidr_blocks=[str(region.network.network_cidr)]
        )
        vpce_sg = vpce_sg_factory.create()

        for interface in vpce or []:
            endpoint = ec2.VpcEndpoint(
                f"com.amazonaws.{region.name}.{interface}",
                vpc_id=vpc.id,
                service_name=f"com.amazonaws.{region.name}.{interface}",
                private_dns_enabled=False,
                vpc_endpoint_type="Interface",
                security_group_ids=[vpce_sg.id],
                subnet_ids=[subnet.id for subnet in subnets.values()],
            )

            endpoint.dns_entries.apply(lambda x: parameter_plain(f"/vpce/{interface}/dns/name", x[0]["dns_name"]))
            endpoint.dns_entries.apply(lambda x: parameter_plain(f"/vpce/{interface}/dns/zone", x[0]["hosted_zone_id"]))

            pulumi.info("Please remember to create relevant DNS entries for Private API endpoints in shared zone")

    pulumi.export("vpc_id", vpc.id)

    return vpc
