import json

import pulumi
from pulumi_aws import (
    securityhub,
    cfg,
    s3,
    iam,
    account,
    ebs,
    get_caller_identity,
    guardduty,
    signer,
    ecr,
    detective,
    wafv2,
)

from cloudlib.aws.ec2 import SecurityGroupFactory
from cloudlib.aws.sns import TopicFactory
from cloudlib.aws.ssm import parameter, parameter_plain
from cloudlib.aws.waf import WebAclFactory
from cloudlib.cloudinfo import Region
from cloudlib.modules.cfg import RecorderFactory, DeliveryChannelFactory
from cloudlib.modules.cloudwatch import LogMetricFilterFactory
from cloudlib.modules.iam import RoleFactory
from cloudlib.modules.kms import KeyFactory


def roles(config, region, stage):
    if region != "us-east-1":
        return

    identity = get_caller_identity()

    # RoleFactory("AWSSupportAccess", assume_role_policy=json.dumps(
    #     {
    #         "Version": "2012-10-17",
    #         "Statement": [
    #             {"Effect": "Allow", "Principal": {"AWS": identity.account_id}}
    #         ],
    #     }
    # ), managed_policy_arns=["arn:aws:iam::aws:policy/AWSSupportAccess"]).create()

    RoleFactory(
        "ExternalAuditor",
        assume_role_policy=json.dumps(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {"AWS": "arn:aws:iam::169677157993:root"},
                        "Action": "sts:AssumeRole",
                        "Condition": {"Bool": {"aws:MultiFactorAuthPresent": "true"}},
                    }
                ],
            }
        ),
        managed_policy_arns=["arn:aws:iam::aws:policy/SecurityAudit", "arn:aws:iam::aws:policy/ReadOnlyAccess"],
    ).create()

    RoleFactory(
        "AWSAppSyncCloudWatchLogsRole",
        assume_role_policy=json.dumps(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {"Effect": "Allow", "Principal": {"Service": "appsync.amazonaws.com"}, "Action": "sts:AssumeRole"}
                ],
            }
        ),
        inline_policies=[
            dict(
                name="log_policy",
                policy=json.dumps(
                    {
                        "Version": "2012-10-17",
                        "Statement": [
                            {
                                "Effect": "Allow",
                                "Action": ["logs:CreateLogGroup", "logs:CreateLogStream", "logs:PutLogEvents"],
                                "Resource": "*",
                            }
                        ],
                    }
                ),
            )
        ],
    ).create()

    for name in ("SNSSuccessFeedback", "SnsFailureFeedback"):
        RoleFactory(
            name,
            assume_role_policy=json.dumps(
                {
                    "Version": "2012-10-17",
                    "Statement": [
                        {"Effect": "Allow", "Principal": {"Service": "sns.amazonaws.com"}, "Action": "sts:AssumeRole"}
                    ],
                }
            ),
            inline_policies=[
                dict(
                    name="delivery_policy",
                    policy=json.dumps(
                        {
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Effect": "Allow",
                                    "Action": [
                                        "logs:CreateLogGroup",
                                        "logs:CreateLogStream",
                                        "logs:PutLogEvents",
                                        "logs:PutMetricFilter",
                                        "logs:PutRetentionPolicy",
                                    ],
                                    "Resource": ["*"],
                                }
                            ],
                        }
                    ),
                )
            ],
        ).create()

    github = iam.OpenIdConnectProvider(
        "github",
        client_id_lists=["sts.amazonaws.com"],
        thumbprint_lists=["6938fd4d98bab03faadb97b34396831e3780aea1", "1c58a3a8518e8759bf075b76b750d4f2df264fcd"],
        url="https://token.actions.githubusercontent.com",
    )

    pulumi = iam.OpenIdConnectProvider(
        "pulumi",
        client_id_lists=["LivaNova"],
        thumbprint_lists=["6938fd4d98bab03faadb97b34396831e3780aea1", "1c58a3a8518e8759bf075b76b750d4f2df264fcd"],
        url="https://api.pulumi.com/oidc",
    )

    RoleFactory(
        "GitHubAccessRole",
        assume_role_policy=json.dumps(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {
                            "Federated": f"arn:aws:iam::{identity.account_id}:oidc-provider/token.actions.githubusercontent.com"
                        },
                        "Action": "sts:AssumeRoleWithWebIdentity",
                        "Condition": {
                            "StringEquals": {
                                "token.actions.githubusercontent.com:aud": "sts.amazonaws.com",
                            },
                            "StringLike": {"token.actions.githubusercontent.com:sub": ["repo:epsyhealth/*", "repo:epsycloud/*"]},
                        },
                    }
                ],
            }
        ),
        managed_policy_arns=["arn:aws:iam::aws:policy/AdministratorAccess"],
    ).create()

    RoleFactory(
        "PulumiAccessRole",
        assume_role_policy=pulumi.arn.apply(
            lambda arn: json.dumps(
                {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Effect": "Allow",
                            "Principal": {"Federated": arn},
                            "Action": "sts:AssumeRoleWithWebIdentity",
                            "Condition": {
                                "StringLike": {
                                    "api.pulumi.com/oidc:aud": "LivaNova",
                                    "api.pulumi.com/oidc:sub": "pulumi:deploy:org:LivaNova:project:Core:*",
                                },
                            },
                        }
                    ],
                }
            )
        ),
        managed_policy_arns=["arn:aws:iam::aws:policy/AdministratorAccess"],
    ).create()


def account_security(config, region, stage):
    if region != "us-east-1":
        return

    s3.AccountPublicAccessBlock(
        "s3_block_public_access",
        block_public_acls=True,
        block_public_policy=True,
        restrict_public_buckets=True,
        ignore_public_acls=True,
    )

    iam.AccountPasswordPolicy(
        "iam_password_policy",
        allow_users_to_change_password=True,
        minimum_password_length=14,
        require_lowercase_characters=True,
        require_numbers=True,
        require_symbols=True,
        require_uppercase_characters=True,
        password_reuse_prevention=24,
        max_password_age=90,
    )

    account.AlternativeContact(
        "account_security_contact",
        alternate_contact_type="SECURITY",
        email_address="security@epsyhealth.com",
        phone_number="+447474994611",
        title="Epsy Security Team",
    )

    ebs.EncryptionByDefault("ebs_enable_encryption", enabled=True)


def account_notifications(config, region, stage):
    identity = get_caller_identity()

    log_group_name = "/cloudtrail/cloud-audit"

    key_factory = KeyFactory("squadcast_alerts")
    key_factory.policy["Statement"].append(
        {
            "Effect": "Allow",
            "Principal": {"Service": f"cloudwatch.amazonaws.com"},
            "Action": ["kms:Decrypt", "kms:GenerateDataKey*"],
            "Resource": "*",
        }
    )
    key_factory.with_alias("squadcast-alerts")

    key = key_factory.create()

    topic_factory = TopicFactory(f"foxglove-{config.stage}-cloudwatch-alerts", kms_master_key_id=key.arn)
    topic_factory.with_https_subscription(
        target="https://api.squadcast.com/v1/incidents/cloudwatch/1ad939691ea252686829254a3575a6a6fe405f56",
        resource_name="squadcast_incidents",
    )
    topic_factory.create()

    topics = {}
    for name, target in config.get_object("sns-sources").items():
        topic_factory = TopicFactory(f"squadcast_alert_{name}", kms_master_key_id=key.arn)
        topic_factory.with_https_subscription(target=target)

        topics[name] = topic_factory.create()

    monitor = [
        dict(
            name="[CloudWatch.2] Ensure a log metric filter and alarm exist for unauthorized API calls",
            pattern='{($.errorCode="*UnauthorizedOperation") || ($.errorCode="AccessDenied*")}',
            filter=dict(name="UnauthorizedApiCalls"),
        ),
        dict(
            name="[CloudWatch.10] Ensure a log metric filter and alarm exist for security group changes",
            pattern="{($.eventName=AuthorizeSecurityGroupIngress) || ($.eventName=AuthorizeSecurityGroupEgress) || "
            + "($.eventName=RevokeSecurityGroupIngress) || ($.eventName=RevokeSecurityGroupEgress) "
            + "|| ($.eventName=CreateSecurityGroup) || ($.eventName=DeleteSecurityGroup)}",
            filter=dict(name="SecurityGroupChanges"),
        ),
        dict(
            name='[CloudWatch.1] A log metric filter and alarm should exist for usage of the "root" user',
            pattern='{$.userIdentity.type="Root" && $.userIdentity.invokedBy NOT EXISTS && $.eventType !="AwsServiceEvent"}',
            filter=dict(name="RootAccountUssage"),
        ),
        dict(
            name="[CloudWatch.3] Ensure a log metric filter and alarm exist for Management Console sign-in without MFA",
            pattern='{ ($.eventName = "ConsoleLogin") && ($.additionalEventData.MFAUsed != "Yes") && ($.userIdentity.type = "IAMUser") && ($.responseElements.ConsoleLogin = "Success") }',
            filter=dict(name="LoginWithoutMFA"),
        ),
        dict(
            name="[CloudWatch.4] Ensure a log metric filter and alarm exist for IAM policy changes",
            pattern="{($.eventSource=iam.amazonaws.com) && (($.eventName=DeleteGroupPolicy) || ($.eventName=DeleteRolePolicy) || ($.eventName=DeleteUserPolicy) || ($.eventName=PutGroupPolicy) || ($.eventName=PutRolePolicy) || ($.eventName=PutUserPolicy) || ($.eventName=CreatePolicy) || ($.eventName=DeletePolicy) || ($.eventName=CreatePolicyVersion) || ($.eventName=DeletePolicyVersion) || ($.eventName=AttachRolePolicy) || ($.eventName=DetachRolePolicy) || ($.eventName=AttachUserPolicy) || ($.eventName=DetachUserPolicy) || ($.eventName=AttachGroupPolicy) || ($.eventName=DetachGroupPolicy))}",
            filter=dict(name="IAMPolicyChanges"),
        ),
        dict(
            name="[CloudWatch.5] Ensure a log metric filter and alarm exist for CloudTrail AWS Configuration changes",
            pattern="{($.eventName=CreateTrail) || ($.eventName=UpdateTrail) || ($.eventName=DeleteTrail) || ($.eventName=StartLogging) || ($.eventName=StopLogging)}",
            filter=dict(name="CloudTrailChanges"),
        ),
        dict(
            name="[CloudWatch.6] Ensure a log metric filter and alarm exist for AWS Management Console authentication failures",
            pattern='{($.eventName=ConsoleLogin) && ($.errorMessage="Failed authentication")}',
            filter=dict(name="ConsoleAuthenticationFailures"),
        ),
        dict(
            name="[CloudWatch.7] Ensure a log metric filter and alarm exist for disabling or scheduled deletion of customer managed keys",
            pattern="{($.eventSource=kms.amazonaws.com) && (($.eventName=DisableKey) || ($.eventName=ScheduleKeyDeletion))}",
            filter=dict(name="KMSDeletion"),
        ),
        dict(
            name="[CloudWatch.8] Ensure a log metric filter and alarm exist for S3 bucket policy changes",
            pattern="{($.eventSource=s3.amazonaws.com) && (($.eventName=PutBucketAcl) || ($.eventName=PutBucketPolicy) || ($.eventName=PutBucketCors) || ($.eventName=PutBucketLifecycle) || ($.eventName=PutBucketReplication) || ($.eventName=DeleteBucketPolicy) || ($.eventName=DeleteBucketCors) || ($.eventName=DeleteBucketLifecycle) || ($.eventName=DeleteBucketReplication))}",
            filter=dict(name="S3BucketChange"),
        ),
        dict(
            name="[CloudWatch.9] Ensure a log metric filter and alarm exist for AWS Config configuration changes",
            pattern="{($.eventSource=config.amazonaws.com) && (($.eventName=StopConfigurationRecorder) || ($.eventName=DeleteDeliveryChannel) || ($.eventName=PutDeliveryChannel) || ($.eventName=PutConfigurationRecorder))}",
            filter=dict(name="AWSConfigChanges"),
        ),
        dict(
            name="[CloudWatch.11] Ensure a log metric filter and alarm exist for changes to Network Access Control Lists (NACL)",
            pattern="{($.eventName=CreateNetworkAcl) || ($.eventName=CreateNetworkAclEntry) || ($.eventName=DeleteNetworkAcl) || ($.eventName=DeleteNetworkAclEntry) || ($.eventName=ReplaceNetworkAclEntry) || ($.eventName=ReplaceNetworkAclAssociation)}",
            filter=dict(name="NACLChanges"),
        ),
        dict(
            name="[CloudWatch.12] Ensure a log metric filter and alarm exist for changes to network gateways",
            pattern="{($.eventName=CreateCustomerGateway) || ($.eventName=DeleteCustomerGateway) || ($.eventName=AttachInternetGateway) || ($.eventName=CreateInternetGateway) || ($.eventName=DeleteInternetGateway) || ($.eventName=DetachInternetGateway)}",
            filter=dict(name="NetworkGatewayChanges"),
        ),
        dict(
            name="[CloudWatch.13] Ensure a log metric filter and alarm exist for route table changes",
            pattern="{($.eventSource=ec2.amazonaws.com) && (($.eventName=CreateRoute) || ($.eventName=CreateRouteTable) || ($.eventName=ReplaceRoute) || ($.eventName=ReplaceRouteTableAssociation) || ($.eventName=DeleteRouteTable) || ($.eventName=DeleteRoute) || ($.eventName=DisassociateRouteTable))}",
            filter=dict(name="RouteTableChanges"),
        ),
        dict(
            name="[CloudWatch.14] Ensure a log metric filter and alarm exist for VPC changes",
            pattern="{($.eventName=CreateVpc) || ($.eventName=DeleteVpc) || ($.eventName=ModifyVpcAttribute) || ($.eventName=AcceptVpcPeeringConnection) || ($.eventName=CreateVpcPeeringConnection) || ($.eventName=DeleteVpcPeeringConnection) || ($.eventName=RejectVpcPeeringConnection) || ($.eventName=AttachClassicLinkVpc) || ($.eventName=DetachClassicLinkVpc) || ($.eventName=DisableVpcClassicLink) || ($.eventName=EnableVpcClassicLink)}",
            filter=dict(name="VPCChanges"),
        ),
        dict(
            name="[IAM.20] Avoid the use of the root user",
            pattern='{$.userIdentity.type="Root" && $.userIdentity.invokedBy NOT EXISTS && $.eventType !="AwsServiceEvent"}',
            filter=dict(name="RootUserUsage"),
        ),
    ]

    for check in monitor:
        filter_factory = LogMetricFilterFactory(
            name=check.get("name"), log_group_name=log_group_name, pattern=check.get("pattern")
        )
        filter_factory.apply(name=check.get("filter").get("name"), namespace="LogMetrics", value=1, default_value=0)
        filter_factory.with_alarm(topics["cloudwatch"].arn)

        filter_factory.create()


def account_security_hub(config, region, stage):
    securityhub.StandardsSubscription(
        "cis_1_2_0",
        standards_arn="arn:aws:securityhub:::ruleset/cis-aws-foundations-benchmark/v/1.2.0",
    )

    securityhub.StandardsSubscription(
        "cis_1_4_0",
        standards_arn="arn:aws:securityhub:us-east-1::standards/cis-aws-foundations-benchmark/v/1.4.0",
    )

    securityhub.StandardsSubscription(
        "nist_5_0)",
        standards_arn="arn:aws:securityhub:us-east-1::standards/nist-800-53/v/5.0.0",
    )

    securityhub.StandardsSubscription(
        "aws_best_practices",
        standards_arn="arn:aws:securityhub:us-east-1::standards/aws-foundational-security-best-practices/v/1.0.0",
    )


def account_security_forwarding(config, region, stage, audit_stack):
    # configure aws config
    recorder_factory = RecorderFactory(
        config=config,
        name="stage-recorder",
    )
    recorder_factory.delivery_factory = DeliveryChannelFactory(
        name="audit-delivery", s3_bucket_name=audit_stack.get_output("config_bucket")
    )
    recorder = recorder_factory.create()

    cfg.AggregateAuthorization(
        "aggregate_authorization",
        account_id=config.get("audit-account"),
        region=region,
        opts=pulumi.ResourceOptions(depends_on=[recorder]),
    )

    # accept invitation to security hub
    if config.get_bool("create_security_hub", config.get_bool("create-security-hub", True)):
        invitee_account = securityhub.Account("account", opts=pulumi.ResourceOptions(depends_on=[recorder]))

    securityhub.InviteAccepter(
        "securityhub_invite_accepter",
        master_id=config.get("audit-account"),
        # opts=pulumi.ResourceOptions(depends_on=[invitee_account]),
    )

    if config.get_bool("create_guardduty_detector", True):
        member_detector = guardduty.Detector("detector")

        guardduty.InviteAccepter(
            "acceptor", detector_id=member_detector.id, master_account_id=config.get("audit-account")
        )

    detective.InvitationAccepter("member", graph_arn=audit_stack.get_output("detective_graph_arn"))


def code_sign_profile():
    profile = signer.SigningProfile("lambda", name_prefix="lambda_", platform_id="AWSLambda-SHA384-ECDSA")
    parameter("/global/lambda-signing-profile", profile.name)
    parameter("/deployments/lambda/signing-profile", profile.name)


def ecr_security():
    ecr.RegistryScanningConfiguration(
        "private_registry_scan_configuration",
        rules=[
            ecr.RegistryScanningConfigurationRuleArgs(
                repository_filters=[
                    ecr.RegistryScanningConfigurationRuleRepositoryFilterArgs(
                        filter="*",
                        filter_type="WILDCARD",
                    )
                ],
                scan_frequency="SCAN_ON_PUSH",
            )
        ],
        scan_type="BASIC",
    )


def security_groups(config: Region, vpc_id):
    lambda_service_factory = SecurityGroupFactory(
        "lambda-service", description="Generic SG for Lambda functions", vpc_id=vpc_id
    )
    lambda_service_factory.with_https_egress([str("0.0.0.0/0")])
    sg = lambda_service_factory.create()
    parameter_plain("/vpc/security-group/lambda-service", value=sg.id)


def waf(audit_stack, name, scope, white_list_ips=None, disable_graphql_schema=False, allow_localhost=True):
    waf_factory = WebAclFactory(f"{name.lower()}-waf", scope=scope.upper())

    if white_list_ips:
        ipset = waf_factory.with_ipset(name=f"[{waf_factory.name}] Whitelist", ip_addresses=white_list_ips)
        waf_factory.with_rule(
            name="Whitelist",
            priority=60,
            statement=wafv2.WebAclRuleStatementArgs(
                ip_set_reference_statement=wafv2.WebAclRuleStatementIpSetReferenceStatementArgs(arn=ipset.arn)
            ),
            visibility_config=wafv2.WebAclVisibilityConfigArgs(
                cloudwatch_metrics_enabled=True, sampled_requests_enabled=True, metric_name="Whitelist"
            ),
            action=wafv2.WebAclRuleActionArgs(allow=wafv2.WebAclRuleActionAllowArgs()),
        )

    if disable_graphql_schema:
        waf_factory.with_rule(
            name="BlockGraphQLSchema",
            priority=50,
            statement=wafv2.WebAclRuleStatementArgs(
                byte_match_statement=wafv2.WebAclRuleStatementByteMatchStatementArgs(
                    positional_constraint="CONTAINS",
                    search_string="__schema",
                    text_transformations=[
                        wafv2.WebAclRuleStatementByteMatchStatementTextTransformationArgs(type="LOWERCASE", priority=0)
                    ],
                    field_to_match=wafv2.WebAclRuleStatementByteMatchStatementFieldToMatchArgs(
                        body=wafv2.WebAclRuleStatementByteMatchStatementFieldToMatchBodyArgs()
                    ),
                )
            ),
            visibility_config=wafv2.WebAclVisibilityConfigArgs(
                cloudwatch_metrics_enabled=True, sampled_requests_enabled=True, metric_name="BlockGraphQLSchema"
            ),
            action=wafv2.WebAclRuleActionArgs(block=wafv2.WebAclRuleActionBlockArgs()),
        )

    waf_factory.with_managed_rule(
        "AWSManagedRulesAmazonIpReputationList", vendor="AWS", rule_name="AWSManagedRulesAmazonIpReputationList"
    )
    waf_factory.with_managed_rule(
        "AWSManagedRulesAnonymousIpList", vendor="AWS", rule_name="AWSManagedRulesAnonymousIpList"
    )

    if scope.upper() != "CLOUDFRONT":
        waf_factory.with_managed_rule(
            "AWSManagedRulesBotControlRuleSet", vendor="AWS", rule_name="AWSManagedRulesBotControlRuleSet"
        )

    waf_factory.with_managed_rule(
        "AWSManagedRulesCommonRuleSet", vendor="AWS", rule_name="AWSManagedRulesCommonRuleSet"
    )

    if allow_localhost:
        waf_factory.with_rule(
            name="AllowDevEnv",
            priority=40,
            statement=wafv2.WebAclRuleStatementArgs(
                or_statement=wafv2.WebAclRuleStatementOrStatementArgs(
                    statements=[
                        wafv2.WebAclRuleStatementArgs(
                            byte_match_statement=wafv2.WebAclRuleStatementByteMatchStatementArgs(
                                positional_constraint="CONTAINS",
                                search_string="localhost",
                                text_transformations=[
                                    wafv2.WebAclRuleStatementByteMatchStatementTextTransformationArgs(
                                        type="LOWERCASE", priority=0
                                    )
                                ],
                                field_to_match=wafv2.WebAclRuleStatementByteMatchStatementFieldToMatchArgs(
                                    single_query_argument=wafv2.WebAclRuleStatementByteMatchStatementFieldToMatchSingleQueryArgumentArgs(
                                        name="redirect_uri"
                                    )
                                ),
                            )
                        ),
                        wafv2.WebAclRuleStatementArgs(
                            byte_match_statement=wafv2.WebAclRuleStatementByteMatchStatementArgs(
                                positional_constraint="CONTAINS",
                                search_string="localhost",
                                text_transformations=[
                                    wafv2.WebAclRuleStatementByteMatchStatementTextTransformationArgs(
                                        type="LOWERCASE", priority=0
                                    )
                                ],
                                field_to_match=wafv2.WebAclRuleStatementByteMatchStatementFieldToMatchArgs(
                                    single_query_argument=wafv2.WebAclRuleStatementByteMatchStatementFieldToMatchSingleQueryArgumentArgs(
                                        name="logout_uri"
                                    )
                                ),
                            )
                        ),
                    ]
                )
            ),
            visibility_config=wafv2.WebAclVisibilityConfigArgs(
                cloudwatch_metrics_enabled=True, sampled_requests_enabled=True, metric_name="AllowDevEnv"
            ),
            action=wafv2.WebAclRuleActionArgs(allow=wafv2.WebAclRuleActionAllowArgs()),
        )

    waf_factory.with_logging(audit_stack.get_output("waf_logs_bucket"))

    # extra config required for rules listed below
    # waf_factory.with_managed_rule("AWSManagedRulesACFPRuleSet", vendor="AWS", rule_name="AWSManagedRulesACFPRuleSet")
    # waf_factory.with_managed_rule("AWSManagedRulesATPRuleSet", vendor="AWS", rule_name="AWSManagedRulesATPRuleSet")

    waf = waf_factory.create()

    # parameter_plain("/security/waf/arn", waf.arn)
    parameter_plain(f"/security/waf/{scope.lower()}/arn", waf.arn)

    return waf
