import json
import re
from os import getenv
from os.path import dirname, join, realpath
from typing import Optional, Callable

import pulumi
from dynaconf import Dynaconf
from pulumi_aws import get_region

from cloudlib.cloudinfo import Region


class Config(pulumi.Config):
    def __init__(self, name: Optional[str] = None) -> None:
        super().__init__(name)

        stack = pulumi.metadata.get_stack()

        vault = getenv("AWS_VAULT")

        env, stack_prefix = vault.split("-", 1)

        if not stack.startswith(env):
            raise Exception(
                f"Looks like calling pulumi with invalid credentials. Stack name: '{stack}', "
                f"AWS profile name env '{env}'"
            )

        project = pulumi.metadata.get_project()
        project = project.replace("_", "-")
        prefix, suffix = project.split("-", maxsplit=1)

        if prefix == "service" and vault != f"{env}-{project}":
            raise Exception(f"For the service deployments AWS profile must be '{env}-{project}'")

        if not project.startswith(stack_prefix):
            raise Exception(
                f"Invalid AWS profile used for the app. AWS profile used for the app '{pulumi.metadata.get_project()}', "
                f"should start with {env}-{prefix}, but '{vault}' was used."
            )

        config_path = realpath(join(dirname(__file__), "..", "..", "config"))

        self.stage = env
        self.app = stack_prefix
        self.region = get_region().id
        self.project = pulumi.metadata.get_project()

        if self.stage != "root":
            self.region_config = Region.get(self.stage, self.region)
            self.network = self.region_config.network

        self.config = Dynaconf(
            envvar_prefix="PULUMI",  # export envvars with `export DYNACONF_FOO=bar`.
            settings_files=[
                join(config_path, "global.yml"),
                join(config_path, "stage_" + self.stage + ".yml"),
                join(config_path, "region_" + self.region + ".yml"),
                join(config_path, self.stage, self.region + ".yml"),
            ],  # Load files in the given order.
        )

    def _get(self, key: str, use: Optional[Callable] = None, instead_of: Optional[Callable] = None) -> Optional[str]:
        config_candidate = super()._get(key, use, instead_of)

        if not config_candidate:
            val = self.config.get(key)

            if isinstance(val, list) or isinstance(val, dict):
                val = json.dumps(val)

            config_candidate = val

        return config_candidate
